@@ -19,8 +19,8 @@ Suffixes;
 Suffixes suffixes[SuffixesNumber] =
 {
   {"", "cat"},
-  {".gz", "gzip -d -c"},
-  {".Z", "gzip -d -c"},
+  {".gz", "gzip -d -q -c"},
+  {".Z", "gzip -d -q -c"},
   {".bz2", "bzip2 -d -c"}};
 
 /*****************************************************************************/
@@ -28,6 +28,100 @@ Suffixes suffixes[SuffixesNumber] =
 char **infopaths = 0;
 int infopathcount = 0;
 
+int
+qsort_cmp (const void *base, const void *compared)
+{
+  char *cbase = ((TagTable *) base)->nodename;
+  char *ccompared = ((TagTable *) compared)->nodename;
+  while (*cbase)
+    {
+      if (*cbase != *ccompared)
+	{
+	  return (int) (*cbase) - (int) (*ccompared);
+	}
+      else
+	{
+	  cbase++;
+	  ccompared++;
+	}
+    }
+  return 0;
+}
+
+FILE *
+dirpage_lookup (char **type, char ***message, long *lines,
+		char *filename, char **first_node)
+{
+#define Type (*type)
+#define Message (*message)
+#define Lines (*lines)
+  FILE *id = 0;
+  int filenamelen = strlen (filename);
+  int goodHit = 0;
+  char name[256];
+  char file[256];
+  int i;
+  id = opendirfile (0);
+  if (!id) 
+    return 0;
+  read_item (id, type, message, lines);
+  for (i = 1; i < Lines; i++)	/* initialize node-links for every line */
+    {
+      if ((Message[i][0] == '*') && (Message[i][1] == ' '))
+	{
+	  char *nameend = strchr (Message[i], ':');
+	  if (nameend)
+	    {
+	      if (*(nameend + 1) != ':')	/* form: `* name:(file)node.' */
+		{
+		  char *filestart = strchr (nameend, '(');
+		  if (filestart)
+		    {
+		      char *fileend = strchr (filestart, ')');
+		      if (fileend)
+			{
+			  char *dot = strchr (fileend, '.');
+			  if (dot)
+			    {
+			      if (strncmp (filename, Message[i] + 2, filenamelen) == 0)
+				{
+				  char *tmp = name;
+				  strncpy (file, filestart + 1, fileend - filestart - 1);
+				  file[fileend - filestart - 1] = 0;
+				  strncpy (name, fileend + 1, dot - fileend - 1);
+				  name[dot - fileend - 1] = 0;
+				  while (*tmp == ' ')
+				    tmp++;
+				  if (strlen (name))
+				    {
+				      *first_node = xmalloc (strlen (tmp) + 1);
+				      strcpy ((*first_node), tmp);
+				    }
+				  fclose (id);
+				  id = 0;
+				  if (!strstr (file, ".info"))
+				    strcat (file, ".info");
+				  id = openinfo (file, 0);
+				  goodHit = 1;
+				}
+			    }
+			}
+		    }
+		}
+	    }
+	}
+    }
+  if (!goodHit)
+    {
+      fclose (id);
+      id = 0;
+    }
+  return id;
+#undef Lines
+#undef Message
+#undef Type
+}
+
 void
 freeitem (char **type, char ***buf, long *lines)
 {
@@ -159,14 +253,14 @@ load_tag_table (char **message, long lines)
   register unsigned int j;
   register char *res;
   int cut = 0;			/* holds the number of corrupt lines */
-  if (strcmp ("(Indirect)", message[1]) == 0)	/*
-						 * if in the first line
-						 * there is a (indirect)
-						 * string, skip that line
-						 * by adding the value of
-						 * is_indirect=1 to all
-						 * message[line] references.
-						 */
+  if (strcasecmp ("(Indirect)", message[1]) == 0)	/*
+							   * if in the first line
+							   * there is a (indirect)
+							   * string, skip that line
+							   * by adding the value of
+							   * is_indirect=1 to all
+							   * message[line] references.
+							 */
     is_indirect = 1;
   tag_table = xmalloc ((lines + 1) * sizeof (TagTable));
   for (i = 1; i < lines - is_indirect; i++)
@@ -207,6 +301,9 @@ load_tag_table (char **message, long lines)
 	cut++;			/* increment the number of corrupt entries */
     }
   TagTableEntries = lines - 1 - is_indirect - cut;
+  FirstNodeOffset = tag_table[1].offset;
+  strcpy (FirstNodeName, tag_table[1].nodename);
+  qsort (&tag_table[1], TagTableEntries, sizeof (TagTable), qsort_cmp);
 }
 int
 seek_indirect (FILE * id)
@@ -224,12 +321,17 @@ seek_indirect (FILE * id)
       while ((input = fgetc (id)) != INFO_TAG)
 	if (input == EOF)
 	  {
+	    if (type)
+	      {
+		xfree (type);
+		type = 0;
+	      }
 	    return 0;
 	  }
       seek_pos = ftell (id) - 2;
       fgetc (id);
       fgets (type, 1024, id);
-      if (strncmp ("Indirect:", type, strlen ("Indirect:")) == 0)
+      if (strncasecmp ("Indirect:", type, strlen ("Indirect:")) == 0)
 	{
 	  finito = 1;
 	}
@@ -274,13 +376,22 @@ seek_tag_table (FILE * id)
 		  mvaddstr (maxy - 1, 0, _ ("Warning: could not find tag table"));
 		  attrset (normal);
 		}
+	      if (type)
+		{
+		  xfree (type);
+		  type = 0;
+		}
 	      return 2;
 	    }
 	}
       seek_pos = ftell (id) - 2;
-      fgetc (id);
+      while (fgetc (id) != '\n')
+	{
+	  if (feof (id))
+	    break;
+	}
       fgets (type, 1024, id);
-      if (strncmp ("Tag Table:", type, strlen ("Tag Table:")) == 0)
+      if (strncasecmp ("Tag Table:", type, strlen ("Tag Table:")) == 0)
 	{
 	  finito = 1;
 	}
@@ -310,6 +421,153 @@ buildcommand (char *dest, char *command, char *filename, const char *tmpfilename
   strcat (dest, tmpfilename);
 }
 
+inline void
+builddircommand (char *dest, char *command, char *filename, const char *tmpfilename)
+{
+  strcpy (dest, command);
+  strcat (dest, " ");
+  strcat (dest, filename);
+  strcat (dest, ">> ");
+  strcat (dest, tmpfilename);
+}
+
+FILE *
+opendirfile (int number)
+{
+  FILE *id = NULL;
+  char buf[1024];		/* holds local copy of filename */
+  char *bufend;			/* points at the trailing 0 of initial name */
+  char command[1128];		/* holds command to evaluate for decompression of file */
+  int i, j;
+  char *tmpfilename;
+  int *fileendentries = xmalloc (infopathcount * sizeof (int));
+  int dir_found = 0;
+  int dircount = 0;
+  int lang_found;
+  struct stat status;
+
+  if (number == 0)		/* initialize tmp filename for file 1 */
+    {
+      if (tmpfilename1)
+	{
+	  unlink (tmpfilename1);	/* erase old tmpfile */
+	  free (tmpfilename1);
+	}
+      tmpfilename1 = tempnam ("/tmp", NULL);
+      tmpfilename = tmpfilename1;	/* later we will refere only to tmp1 */
+    }
+  for (i = 0; i < infopathcount; i++)	/* go through all paths */
+    {
+      lang_found = 0;
+      strcpy (buf, infopaths[i]);	/* build a filename */
+      strcat (buf, "/");
+      if (getenv ("LANG") != NULL)
+	strcat (buf, getenv ("LANG"));
+      strcat (buf, "/dir");
+      bufend = buf;		/* 
+				 * remember the bufend to make it
+				 * possible later to glue compression
+				 * suffixes.
+				 */
+      bufend += strlen (buf);
+      for (j = 0; j < SuffixesNumber; j++)	/* go through all suffixes */
+	{
+	  strcat (buf, suffixes[j].suffix);
+	  if ((id = fopen (buf, "r")) != NULL)
+	    {
+	      fclose (id);
+	      builddircommand (command, suffixes[j].command, buf, tmpfilename);
+	      system (command);
+	      lstat (tmpfilename, &status);
+	      fileendentries[dircount] = status.st_size;
+	      dircount++;
+	      dir_found = 1;
+	      lang_found = 1;
+	    }
+	  (*bufend) = 0;
+	}
+
+      /* same as above, but without $LANG support */
+      if (!lang_found)
+	{
+	  strcpy (buf, infopaths[i]);	/* build a filename */
+	  strcat (buf, "/");
+	  strcat (buf, "dir");
+	  bufend = buf;		/* 
+				 * remember the bufend to make it
+				 * possible later to glue compression
+				 * suffixes.
+				 */
+	  bufend += strlen (buf);
+	  for (j = 0; j < SuffixesNumber; j++)	/* go through all suffixes */
+	    {
+	      strcat (buf, suffixes[j].suffix);
+	      if ((id = fopen (buf, "r")) != NULL)
+		{
+		  fclose (id);
+		  builddircommand (command, suffixes[j].command, buf, tmpfilename);
+		  system (command);
+		  lstat (tmpfilename, &status);
+		  fileendentries[dircount] = status.st_size;
+		  dircount++;
+		  dir_found = 1;
+		}
+	      (*bufend) = 0;
+	    }
+	}
+    }
+  if (dir_found)
+    id = fopen (tmpfilename, "r");
+  if (id)	/*
+                 * Filter the concatenated dir pages to exclude hidden
+                 * parts of info entries
+                 */
+    {
+      char *tmp;
+      long filelen, i;
+      int aswitch = 0;
+      int firstswitch = 0;
+      dircount = 0;
+
+      fseek (id, 0, SEEK_END);
+      filelen = ftell (id);
+
+      tmp = xmalloc (filelen);
+      fseek (id, 0, SEEK_SET);
+      fread (tmp, 1, filelen, id);
+      fclose (id);
+      id = fopen (tmpfilename, "w");
+      for (i = 0; i < filelen; i++)
+	{
+	  if (tmp[i] == INFO_TAG)
+	    {
+	      aswitch ^= 1;
+	      if (!firstswitch)
+		fputc (tmp[i], id);
+	      firstswitch = 1;
+	    }
+	  else if ((aswitch) || (!firstswitch))
+	    fputc (tmp[i], id);
+	  if (i + 1 == fileendentries[dircount])
+	    {
+	      if (aswitch != 0)
+		aswitch = 0;
+	      dircount++;	/* the last dircount should fit to the end of filelen */
+	    }
+	}
+      fputc (INFO_TAG, id);
+      fputc ('\n', id);
+      xfree (fileendentries);
+      fclose (id);
+      id = fopen (tmpfilename, "r");
+      xfree (tmp);
+
+      return id;
+    }
+  return NULL;
+}
+
+
 FILE *
 openinfo (char *filename, int number)	/*
 					 * Note: openinfo is a function
@@ -334,6 +592,9 @@ openinfo (char *filename, int number)	/*
   int i, j, twoloops;
   char *tmpfilename;
 
+  if (strncmp (filename, "dir", 3) == 0)
+    return opendirfile (number);
+
   if (number == 0)		/* initialize tmp filename for file 1 */
     {
       if (tmpfilename1)
@@ -403,13 +664,18 @@ openinfo (char *filename, int number)	/*
 void
 initpaths ()
 {
+  char curdir[1024];
   int i, infolen, wsklen;
+  int localdir = 0, globaldir = 0;	/* flags, which say if the /usr/local/info *
+					 * and /usr/info were defined in $INFOPATH */
+  int sharedir = 0;		/* the same for /usr/share/info */
   char *lang = getenv ("LANG");
+
+  getcwd (curdir, 1024);
+
   infopaths = xmalloc (5 * sizeof (char *));
 
-  infopaths[0] = xmalloc (5);
-  strcpy (infopaths[0], ".");	/* for ./ */
-  infopathcount = 1;
+  infopathcount = 0;
 
   if (getenv ("INFOPATH") != NULL)	/* check paths in $INFOPATH variable */
     {
@@ -428,6 +694,12 @@ initpaths ()
 	{
 	  infopaths = xrealloc (infopaths, (infopathcount + 2) * sizeof (char *));
 	  wsklen = strlen (wsk);
+	  if (strncmp ("/usr/info", wsk, 9) == 0)
+	    globaldir = 1;
+	  if (strncmp ("/usr/local/info", wsk, 15) == 0)
+	    localdir = 1;
+	  if (strncmp ("/usr/share/info", wsk, 15) == 0)
+	    sharedir = 1;
 	  if (wsklen)
 	    {
 	      if (wsk[wsklen - 1] == '/')
@@ -435,7 +707,7 @@ initpaths ()
 	    }
 	  if (lang)
 	    {
-	      infopaths[infopathcount] = xmalloc (wsklen + strlen (lang) + 1);
+	      infopaths[infopathcount] = xmalloc (wsklen + strlen (lang) + 2);
 	      strcpy (infopaths[infopathcount], wsk);
 	      strcat (infopaths[infopathcount], "/");
 	      strcat (infopaths[infopathcount], lang);
@@ -461,59 +733,96 @@ initpaths ()
   if (lang)
     {
       infopaths = xrealloc (infopaths, sizeof (char *) * (infopathcount + 15));
-      infopaths[infopathcount] = xmalloc (15 + strlen (lang));
-      strcpy (infopaths[infopathcount], "/usr/info/");
-      strcat (infopaths[infopathcount], lang);
-      infopathcount++;
-      infopaths[infopathcount] = xmalloc (25 + strlen (lang));
-      strcpy (infopaths[infopathcount], "/usr/local/info/");
-      strcat (infopaths[infopathcount], lang);
-      infopathcount++;
-      infopaths[infopathcount] = xmalloc (25 + strlen (lang));
-      strcpy (infopaths[infopathcount], "/usr/share/info/");
-      strcat (infopaths[infopathcount], lang);
-      infopathcount++;
-      infopaths[infopathcount] = xmalloc (15);
-      strcpy (infopaths[infopathcount], "/usr/info");
-      infopathcount++;
-      infopaths[infopathcount] = xmalloc (25);
-      strcpy (infopaths[infopathcount], "/usr/local/info");
-      infopathcount++;
-      infopaths[infopathcount] = xmalloc (25);
-      strcpy (infopaths[infopathcount], "/usr/share/info");
-      infopathcount++;
+      if ((!globaldir) && (strcmp (curdir, "/usr/info")))
+	{
+	  infopaths[infopathcount] = xmalloc (15 + strlen (lang));
+	  strcpy (infopaths[infopathcount], "/usr/info");
+	  strcat (infopaths[infopathcount], lang);
+	  infopathcount++;
+	}
+      if ((!localdir) && (strcmp (curdir, "/usr/local/info")))
+	{
+	  infopaths[infopathcount] = xmalloc (25 + strlen (lang));
+	  strcpy (infopaths[infopathcount], "/usr/local/info");
+	  strcat (infopaths[infopathcount], lang);
+	  infopathcount++;
+	}
+      if ((!sharedir) && (strcmp (curdir, "/usr/share/info")))
+	{
+	  infopaths[infopathcount] = xmalloc (25 + strlen (lang));
+	  strcpy (infopaths[infopathcount], "/usr/share/info");
+	  strcat (infopaths[infopathcount], lang);
+	  infopathcount++;
+	}
+      if ((!globaldir) && (strcmp (curdir, "/usr/info")))
+	{
+	  infopaths[infopathcount] = xmalloc (15);
+	  strcpy (infopaths[infopathcount], "/usr/info");
+	  infopathcount++;
+	}
+      if ((!localdir) && (strcmp (curdir, "/usr/local/info")))
+	{
+	  infopaths[infopathcount] = xmalloc (25);
+	  strcpy (infopaths[infopathcount], "/usr/local/info");
+	  infopathcount++;
+	}
+      if ((!sharedir) && (strcmp (curdir, "/usr/share/info")))
+	{
+	  infopaths[infopathcount] = xmalloc (25);
+	  strcpy (infopaths[infopathcount], "/usr/share/info");
+	  infopathcount++;
+	}
     }
   else
     {
       infopaths = xrealloc (infopaths, sizeof (char *) * (infopathcount + 15));
-      infopaths[infopathcount] = xmalloc (15);
-      strcpy (infopaths[infopathcount], "/usr/info");
-      infopathcount++;
-      infopaths[infopathcount] = xmalloc (25);
-      strcpy (infopaths[infopathcount], "/usr/local/info");
-      infopathcount++;
-      infopaths[infopathcount] = xmalloc (25);
-      strcpy (infopaths[infopathcount], "/usr/share/info");
-      infopathcount++;
+      if ((!globaldir) && (strcmp (curdir, "/usr/info")))
+	{
+	  infopaths[infopathcount] = xmalloc (15);
+	  strcpy (infopaths[infopathcount], "/usr/info");
+	  infopathcount++;
+	}
+      if ((!localdir) && (strcmp (curdir, "/usr/local/info")))
+	{
+	  infopaths[infopathcount] = xmalloc (25);
+	  strcpy (infopaths[infopathcount], "/usr/local/info");
+	  infopathcount++;
+	}
+      if ((!sharedir) && (strcmp (curdir, "/usr/share/info")))
+	{
+	  infopaths[infopathcount] = xmalloc (25);
+	  strcpy (infopaths[infopathcount], "/usr/share/info");
+	  infopathcount++;
+	}
     }
+  infopaths[infopathcount] = xmalloc (5);
+  strcpy (infopaths[infopathcount], ".");	/* for ./ */
+  infopathcount++;
 }
 void
 create_indirect_tag_table ()
 {
-  FILE *id;
+  FILE *id = 0;
   int i, j, initial;
   for (i = 1; i <= IndirectEntries; i++)
     {
       id = openinfo (indirect[i].filename, 1);
       initial = TagTableEntries + 1;
       if (id)
-	create_tag_table (id);
+	{
+	  create_tag_table (id);
+	  FirstNodeOffset = tag_table[1].offset;
+	  strcpy (FirstNodeName, tag_table[1].nodename);
+	}
       fclose (id);
       for (j = initial; j <= TagTableEntries; j++)
 	{
-	  tag_table[j].offset += (indirect[i].offset - tag_table[1].offset);
+	  tag_table[j].offset += (indirect[i].offset - FirstNodeOffset);
 	}
     }
+  FirstNodeOffset = tag_table[1].offset;
+  strcpy (FirstNodeName, tag_table[1].nodename);
+  qsort (&tag_table[1], TagTableEntries, sizeof (TagTable), qsort_cmp);
 }
 void
 create_tag_table (FILE * id)
@@ -549,7 +858,8 @@ create_tag_table (FILE * id)
 	  else
 	    {
 	      int colons = 0, i, j;
-	      for (i = 0; i < strlen (buf); i++)
+	      int buflen = strlen (buf);
+	      for (i = 0; i < buflen; i++)
 		{
 		  if (buf[i] == ':')
 		    colons++;
@@ -561,12 +871,13 @@ create_tag_table (FILE * id)
 					 * bellow.
 					 */
 		    {
-		      for (j = i + 2; j < strlen (buf); j++)
+		      for (j = i + 2; j < buflen; j++)
 			{
 			  if ((buf[j] == ',') || (buf[j] == '\n'))
 			    {
 			      tag_table = xrealloc (tag_table, sizeof (TagTable) * (TagTableEntries + 1));
 			      buf[j] = 0;
+			      buflen = j;
 			      strcpy (tag_table[TagTableEntries].nodename, buf + i + 2);
 			      tag_table[TagTableEntries].offset = oldpos - 2;
 			      break;
@@ -580,6 +891,12 @@ create_tag_table (FILE * id)
     }				/* end: global while loop, looping until eof */
   xfree (buf);
   buf = 0;
+  if (!indirect)
+    {
+      FirstNodeOffset = tag_table[1].offset;
+      strcpy (FirstNodeName, tag_table[1].nodename);
+      qsort (&tag_table[1], TagTableEntries, sizeof (TagTable), qsort_cmp);
+    }
 }
 
 void
@@ -597,6 +914,7 @@ seeknode (int tag_table_pos, FILE ** Id)
 	{
 	  if (indirect[i].offset <= tag_table[tag_table_pos].offset)
 	    {
+	      long off = tag_table[tag_table_pos].offset - indirect[i].offset + FirstNodeOffset - 4;
 	      fclose (id);
 	      id = openinfo (indirect[i].filename, 0);
 	      if (id == NULL)
@@ -605,14 +923,39 @@ seeknode (int tag_table_pos, FILE ** Id)
 		  printf (_ ("Error: could not open info file\n"));
 		  exit (1);
 		}
-	      fseek (id, tag_table[tag_table_pos].offset - indirect[i].offset + tag_table[1].offset - 2, SEEK_SET);
+	      if (off > 0)
+		fseek (id, off, SEEK_SET);
+	      else
+		fseek (id, off, SEEK_SET);
 	      break;
 	    }
 	}
     }
   else
     {
-      fseek (id, tag_table[tag_table_pos].offset - 2, SEEK_SET);
+      long off = tag_table[tag_table_pos].offset - 4;
+      if (off > 0)
+	fseek (id, off, SEEK_SET);
+      else
+	fseek (id, off, SEEK_SET);
     }
 #undef id
 }
+
+void 
+strip_compression_suffix (char *file)	/* removes trailing .gz, .bz2, etc. */
+{
+  char *found = 0;
+  int j;
+  for (j = 0; j < SuffixesNumber; j++)
+    {
+      if (found = strstr (file, suffixes[j].suffix))
+	{
+	  if (*(found + strlen (suffixes[j].suffix)) == 0)
+	    {
+	      *found = 0;
+	      break;
+	    }
+	}
+    }
+}
