@@ -3,18 +3,69 @@
 #define MENU_DOT 0
 #define NOTE_DOT 1
 
+int
+compare_tag_table_string (char *base, char *compared)
+/*
+ * Compares two strings, ignoring whitespaces (tabs, spaces)
+ */
+{
+  int i, j;
+
+  j = 0;
+
+  for (i = 0; base[i] != 0; i++)
+    {
+      if (base[i] != compared[j])
+	{
+	  if (((compared[j] != ' ') && (compared[j] != '\t')) ||
+	      ((base[i] != ' ') && (base[i] != '\t')))
+	    return (int) base[i] - (int) compared[j];
+	  i--;
+	}
+      j++;
+    }
+  while (compared[j])
+    {
+      if (((compared[j] != ' ') && (compared[j] != '\t')) ||
+	  ((base[i] != ' ') && (base[i] != '\t')))
+	return (int) base[i] - (int) compared[j];
+      j++;
+    }
+  return 0;
+}
+
+inline int
+exists_in_tag_table (char *item)	/* checks if an item belongs to
+					 * tag table. returns 1 on success
+					 * and 0 on failure.
+					 * It should be optimised...
+					 */
+{
+  if (gettagtablepos (item) != -1)
+    return 1;
+  else
+    return 0;
+}
+
 int
 calculate_len (char *start, char *end)	/*
 					 * calculates the length of string
 					 * between start and end, counting
-					 * `\t' as 8 chars.
+					 * `\t' as filling up to 8 chars.
+					 * (i.e. at line 22 tab will increment
+					 * the counter by 2 
+					 * [8-(22-int(22/8)*8)] spaces)
 					 */
 {
-  int len = (int) (end - start);
-  while (start <= end)
+  int len = 0;
+  while (start < end)
     {
+      len++;
       if (*start == '\t')
-	len += 7;
+	{
+	  len--;
+	  len += (8 - ((len) - (((len) >> 3) << 3)));
+	}
       start++;
     }
   return len;
@@ -79,7 +130,8 @@ finddot (char *str, int note)	/*
     closest = end[2];
   else if (end[3])
     closest = end[3];
-  for (i = 1; i < note; i++)
+  for (i = 1; i < note; i++)	/* find the delimiter, which was found most
+				   recently */
     {
       if ((end[i] < closest) && (end[i]))
 	closest = end[i];
@@ -123,11 +175,10 @@ void
 initializelinks (char *line1, char *line2, int line)
 {
   char *tmp;
-  char *notestart, *urlstart, *urlend;
-  char *quotestart, *quoteend;
+  char *notestart = 0, *urlstart = 0, *urlend = 0;
+  char *quotestart = 0, *quoteend = 0;
   char *buf = xmalloc (strlen (line1) + strlen (line2) + 1);
-  int i, changed;
-  int hobjectcnt = hyperobjectcount;
+  int changed;
   int line1len = strlen (line1);
   strcpy (buf, line1);		/* copy two lines into one */
   if (strlen (line1))
@@ -142,7 +193,7 @@ initializelinks (char *line1, char *line2, int line)
       changed = 0;
       if ((quotestart = strchr (quoteend, '`')) != NULL)	/* find start of quoted text */
 	{
-	  if (quotestart < buf + line1len)
+	  if (quotestart < buf + line1len)	/* if it's in the first line of the two glued together */
 	    if ((quoteend = strchr (quotestart, '\'')) != NULL)		/* find the end of quoted text */
 	      {
 		if (quoteend - quotestart > 1)
@@ -161,7 +212,7 @@ initializelinks (char *line1, char *line2, int line)
 			else
 			  hyperobjects = xrealloc (hyperobjects, sizeof (HyperObject) * (hyperobjectcount + 1));
 			hyperobjects[hyperobjectcount].line = line;
-			hyperobjects[hyperobjectcount].row = calculate_len (buf, quotestart + 1);
+			hyperobjects[hyperobjectcount].col = calculate_len (buf, quotestart + 1);
 			hyperobjects[hyperobjectcount].breakpos = -1;	/* default */
 			if (quoteend > buf + line1len)
 			  {
@@ -194,7 +245,7 @@ initializelinks (char *line1, char *line2, int line)
 	  else
 	    hyperobjects = xrealloc (hyperobjects, sizeof (HyperObject) * (hyperobjectcount + 1));
 	  hyperobjects[hyperobjectcount].line = line;
-	  hyperobjects[hyperobjectcount].row = calculate_len (line1, urlstart);
+	  hyperobjects[hyperobjectcount].col = calculate_len (line1, urlstart);
 	  hyperobjects[hyperobjectcount].breakpos = -1;
 	  hyperobjects[hyperobjectcount].type = 6;
 	  strncpy (hyperobjects[hyperobjectcount].node, urlstart, urlend - urlstart);
@@ -246,7 +297,7 @@ initializelinks (char *line1, char *line2, int line)
 		  hyperobjects[hyperobjectcount].node[NodenameLen] = 0;
 		  hyperobjects[hyperobjectcount].type = 0;
 		  hyperobjects[hyperobjectcount].line = line;
-		  hyperobjects[hyperobjectcount].row = 2;
+		  hyperobjects[hyperobjectcount].col = 2;
 		  hyperobjects[hyperobjectcount].breakpos = -1;
 		  hyperobjectcount++;
 		}
@@ -255,21 +306,26 @@ initializelinks (char *line1, char *line2, int line)
 	    /* if not cross-info link */
 	    {
 	      long NodenameLen = (long) (tmp - line1 - 2);
+	      int goodHit = 0;
 	      strcpy (hyperobjects[hyperobjectcount].file, "");
 	      strncpy (hyperobjects[hyperobjectcount].node,
 		       line1 + 2, NodenameLen);
 	      hyperobjects[hyperobjectcount].node[NodenameLen] = 0;
 	      hyperobjects[hyperobjectcount].type = 0;
 	      hyperobjects[hyperobjectcount].line = line;
-	      hyperobjects[hyperobjectcount].row = 2;
+	      hyperobjects[hyperobjectcount].col = 2;
 	      hyperobjects[hyperobjectcount].breakpos = -1;
-	      for (i = 1; i <= TagTableEntries; i++)
+	      if (exists_in_tag_table (hyperobjects[hyperobjectcount].node))
 		{
-		  if (strcmp (tag_table[i].nodename,
-			      hyperobjects[hyperobjectcount].node) == 0)
+		  hyperobjectcount++;	/* yep, this was a good hit */
+		  goodHit = 1;
+		}
+	      if (!goodHit)
+		{
+		  if (!hyperobjectcount)
 		    {
-		      hyperobjectcount++;	/* yep, this was a good hit */
-		      break;	/* we can increase the objectcount */
+		      xfree (hyperobjects);
+		      hyperobjects = 0;
 		    }
 		}
 	    }
@@ -309,7 +365,7 @@ initializelinks (char *line1, char *line2, int line)
 		      hyperobjects[hyperobjectcount].node[NodenameLen] = 0;
 		      hyperobjects[hyperobjectcount].type = 1;
 		      hyperobjects[hyperobjectcount].line = line;
-		      hyperobjects[hyperobjectcount].row = calculate_len (line1, start);
+		      hyperobjects[hyperobjectcount].col = calculate_len (line1, start);
 		      hyperobjects[hyperobjectcount].breakpos = -1;
 		      hyperobjectcount++;
 		    }
@@ -324,6 +380,7 @@ initializelinks (char *line1, char *line2, int line)
 	    handle_no_file_menu_label:
 	      {
 		long NodenameLen;
+		int goodHit = 0;	/* has val of 1, if it's a good hit */
 		if (!hyperobjectcount)
 		  hyperobjects = xmalloc (sizeof (HyperObject));
 		else
@@ -339,15 +396,19 @@ initializelinks (char *line1, char *line2, int line)
 		hyperobjects[hyperobjectcount].node[NodenameLen] = 0;
 		hyperobjects[hyperobjectcount].type = 1;
 		hyperobjects[hyperobjectcount].line = line;
-		hyperobjects[hyperobjectcount].row = calculate_len (line1, start);
+		hyperobjects[hyperobjectcount].col = calculate_len (line1, start);
 		hyperobjects[hyperobjectcount].breakpos = -1;
-		for (i = 1; i <= TagTableEntries; i++)
+		if (exists_in_tag_table (hyperobjects[hyperobjectcount].node))
 		  {
-		    if (strcmp (tag_table[i].nodename,
-				hyperobjects[hyperobjectcount].node) == 0)
+		    hyperobjectcount++;		/* yep, this was a good hit */
+		    goodHit = 1;
+		  }
+		if (!goodHit)
+		  {
+		    if (!hyperobjectcount)
 		      {
-			hyperobjectcount++;	/* yep, this was a good hit */
-			break;	/* we can increase the objectcount */
+			xfree (hyperobjects);
+			hyperobjects = 0;
 		      }
 		  }
 	      }
@@ -393,20 +454,20 @@ initializelinks (char *line1, char *line2, int line)
 		      if (notestart + 7 - buf < strlen (line1))
 			{
 			  hyperobjects[hyperobjectcount].line = line;
-			  hyperobjects[hyperobjectcount].row = calculate_len (buf, notestart + 7);
+			  hyperobjects[hyperobjectcount].col = calculate_len (buf, notestart + 7);
 			  if (tmp - buf < strlen (line1))	/* if the note highlight fits int first line */
 			    hyperobjects[hyperobjectcount].breakpos = -1;	/* we don't need to break highlighting int several lines */
 			  else
-			    hyperobjects[hyperobjectcount].breakpos = strlen (line1) - (long) (notestart + 7 - buf);	/* otherwise we need it */
+			    hyperobjects[hyperobjectcount].breakpos = strlen (line1) - (long) (notestart + 7 - buf) + 1;	/* otherwise we need it */
 			}
 		      else
 			{
 			  hyperobjects[hyperobjectcount].line = line + 1;
-			  hyperobjects[hyperobjectcount].row = calculate_len (buf + strlen (line1), notestart + 7);
+			  hyperobjects[hyperobjectcount].col = calculate_len (buf + strlen (line1), notestart + 7);
 			  if (tmp - buf < strlen (line1))	/* as above */
 			    hyperobjects[hyperobjectcount].breakpos = -1;
-			  else
-			    hyperobjects[hyperobjectcount].breakpos = strlen (line1) - (long) (notestart + 7 - buf);
+			  else if ((hyperobjects[hyperobjectcount].breakpos = strlen (line1) - (long) (notestart + 7 - buf) + 1) == 0)
+			    hyperobjects[hyperobjectcount].line--;
 			}
 		      hyperobjectcount++;
 		    }
@@ -415,6 +476,7 @@ initializelinks (char *line1, char *line2, int line)
 		/* if not cross-info link */
 		{
 		  long NodenameLen = (long) (tmp - notestart - 6);
+		  int goodHit = 0;
 		  strcpy (hyperobjects[hyperobjectcount].file, "");
 		  strncpy (hyperobjects[hyperobjectcount].node,
 			   notestart + 6, NodenameLen);
@@ -423,28 +485,32 @@ initializelinks (char *line1, char *line2, int line)
 		  if (notestart + 7 - buf < strlen (line1))
 		    {
 		      hyperobjects[hyperobjectcount].line = line;
-		      hyperobjects[hyperobjectcount].row = calculate_len (buf, notestart + 7) - 1;
+		      hyperobjects[hyperobjectcount].col = calculate_len (buf, notestart + 7) - 1;
 		      if (tmp - buf < strlen (line1))	/* if the note highlight fits int first line */
 			hyperobjects[hyperobjectcount].breakpos = -1;	/* we don't need to break highlighting int several lines */
 		      else
-			hyperobjects[hyperobjectcount].breakpos = strlen (line1) - (long) (notestart + 7 - buf);	/* otherwise we need it */
+			hyperobjects[hyperobjectcount].breakpos = strlen (line1) - (long) (notestart + 7 - buf) + 1;	/* otherwise we need it */
 		    }
 		  else
 		    {
 		      hyperobjects[hyperobjectcount].line = line + 1;
-		      hyperobjects[hyperobjectcount].row = calculate_len (buf + strlen (line1), notestart + 7) - 1;
+		      hyperobjects[hyperobjectcount].col = calculate_len (buf + strlen (line1), notestart + 7) - 1;
 		      if (tmp - buf < strlen (line1))	/* as above */
 			hyperobjects[hyperobjectcount].breakpos = -1;
-		      else
-			hyperobjects[hyperobjectcount].breakpos = strlen (line1) - (long) (notestart + 7 - buf);
+		      else if ((hyperobjects[hyperobjectcount].breakpos = strlen (line1) - (long) (notestart + 7 - buf) + 1) == 0)
+			hyperobjects[hyperobjectcount].line--;
 		    }
-		  for (i = 1; i <= TagTableEntries; i++)
+		  if (exists_in_tag_table (hyperobjects[hyperobjectcount].node))
 		    {
-		      if (strstr (tag_table[i].nodename,
-			       hyperobjects[hyperobjectcount].node) != NULL)
+		      hyperobjectcount++;	/* yep, this was a good hit */
+		      goodHit = 1;
+		    }
+		  if (!goodHit)
+		    {
+		      if (!hyperobjectcount)
 			{
-			  hyperobjectcount++;	/* yep, this was a good hit */
-			  break;	/* we can increase the objectcount */
+			  xfree (hyperobjects);
+			  hyperobjects = 0;
 			}
 		    }
 		}
@@ -486,7 +552,7 @@ initializelinks (char *line1, char *line2, int line)
 			  if (start - buf < strlen (line1))
 			    {
 			      hyperobjects[hyperobjectcount].line = line;
-			      hyperobjects[hyperobjectcount].row = calculate_len (buf, start);
+			      hyperobjects[hyperobjectcount].col = calculate_len (buf, start);
 			      if (dot - buf < strlen (line1))	/* if the note highlight fits in first line */
 				hyperobjects[hyperobjectcount].breakpos = -1;	/* we don't need to break highlighting int several lines */
 			      else
@@ -495,7 +561,7 @@ initializelinks (char *line1, char *line2, int line)
 			  else
 			    {
 			      hyperobjects[hyperobjectcount].line = line + 1;
-			      hyperobjects[hyperobjectcount].row = calculate_len (buf + strlen (line1), start);
+			      hyperobjects[hyperobjectcount].col = calculate_len (buf + strlen (line1), start);
 			      hyperobjects[hyperobjectcount].breakpos = -1;
 			    }
 			  hyperobjectcount++;
@@ -511,6 +577,7 @@ initializelinks (char *line1, char *line2, int line)
 		handle_no_file_note_label:
 		  {
 		    long NodenameLen;
+		    int goodHit = 0;
 		    if (!hyperobjectcount)
 		      hyperobjects = xmalloc (sizeof (HyperObject));
 		    else
@@ -528,7 +595,7 @@ initializelinks (char *line1, char *line2, int line)
 		    if (start - buf < strlen (line1))
 		      {
 			hyperobjects[hyperobjectcount].line = line;
-			hyperobjects[hyperobjectcount].row = calculate_len (buf, start);
+			hyperobjects[hyperobjectcount].col = calculate_len (buf, start);
 			if (dot - buf < strlen (line1))		/* if the note highlight fits in first line */
 			  hyperobjects[hyperobjectcount].breakpos = -1;		/* we don't need to break highlighting int several lines */
 			else
@@ -537,16 +604,20 @@ initializelinks (char *line1, char *line2, int line)
 		    else
 		      {
 			hyperobjects[hyperobjectcount].line = line + 1;
-			hyperobjects[hyperobjectcount].row = calculate_len (strlen (line1) + buf, start);
+			hyperobjects[hyperobjectcount].col = calculate_len (strlen (line1) + buf, start);
 			hyperobjects[hyperobjectcount].breakpos = -1;
 		      }
-		    for (i = 1; i <= TagTableEntries; i++)
+		    if (exists_in_tag_table (hyperobjects[hyperobjectcount].node))
 		      {
-			if (strstr (tag_table[i].nodename,
-			       hyperobjects[hyperobjectcount].node) != NULL)
+			hyperobjectcount++;	/* yep, this was a good hit */
+			goodHit = 1;
+		      }
+		    if (!goodHit)
+		      {
+			if (!hyperobjectcount)
 			  {
-			    hyperobjectcount++;		/* yep, this was a good hit */
-			    break;	/* we can increase the objectcount */
+			    xfree (hyperobjects);
+			    hyperobjects = 0;
 			  }
 		      }
 		  }
@@ -579,7 +650,7 @@ initializelinks (char *line1, char *line2, int line)
       else
 	hyperobjects = xrealloc (hyperobjects, sizeof (HyperObject) * (hyperobjectcount + 1));
       hyperobjects[hyperobjectcount].line = line;
-      hyperobjects[hyperobjectcount].row = calculate_len (line1, urlstart);
+      hyperobjects[hyperobjectcount].col = calculate_len (line1, urlstart);
       hyperobjects[hyperobjectcount].breakpos = -1;
       hyperobjects[hyperobjectcount].type = 4;
       strncpy (hyperobjects[hyperobjectcount].node, urlstart, urlend - urlstart);
@@ -597,7 +668,7 @@ initializelinks (char *line1, char *line2, int line)
       else
 	hyperobjects = xrealloc (hyperobjects, sizeof (HyperObject) * (hyperobjectcount + 1));
       hyperobjects[hyperobjectcount].line = line;
-      hyperobjects[hyperobjectcount].row = calculate_len (line1, urlstart);
+      hyperobjects[hyperobjectcount].col = calculate_len (line1, urlstart);
       hyperobjects[hyperobjectcount].breakpos = -1;
       hyperobjects[hyperobjectcount].type = 5;
       strncpy (hyperobjects[hyperobjectcount].node, urlstart, urlend - urlstart);
