@@ -1,5 +1,5 @@
 #include "common_includes.h"
-#include <regex.h>
+#include <ctype.h>
 
 #ifndef MIN
 #define        MIN(a,b) (((a)<(b))?(a):(b))
@@ -24,8 +24,10 @@ work (char ***message, char **type, long *lines, FILE * id, int tag_table_pos)
   int indirectstart = -1;
   int cursorchanged = 0;
   int key = 0;
+  int highlight_searched = 0;
   int return_value;
   char *token, *tmp;
+  char searchword[256];
 
   if (rval.file)		/* if the static variable was allocated, free it */
     {
@@ -64,8 +66,11 @@ work (char ***message, char **type, long *lines, FILE * id, int tag_table_pos)
 				 * 's'earch function.
 				 */
     {
+      highlight_searched = 1;
+      copy_stripped_from_regexp (searchagain.lastsearch, searchword);
       pos = aftersearch;	/* set pos to the found position */
       aftersearch = 0;
+
     }
 
   if (ncursor != -1)
@@ -77,7 +82,7 @@ work (char ***message, char **type, long *lines, FILE * id, int tag_table_pos)
   npos = -1;			/* turn off the `next-time' pos/cursor modifiers */
   ncursor = -1;
   addtopline (Type);
-  while (key != keys.quit_1)
+  while (1)
     {
       nodelay (stdscr, TRUE);	/* 
 				 * read key, and show screen only if there
@@ -88,6 +93,20 @@ work (char ***message, char **type, long *lines, FILE * id, int tag_table_pos)
       if (key == ERR)
 	{
 	  showscreen (Message, Type, Lines, pos, cursor);
+	  if (highlight_searched)
+	    {
+	      char *highlightstart = strstr ((*message)[pos], searchword);
+	      attrset (searchhighlight);
+	      if (highlightstart)
+		{
+		  mvaddstr (1,
+			    calculate_len ((*message)[pos], highlightstart),
+			    searchword);
+		}
+	      highlight_searched = 0;
+	      attrset (normal);
+	      move (0, 0);
+	    }
 	  nodelay (stdscr, FALSE);
 	  key = pinfo_getch ();
 	}
@@ -97,6 +116,53 @@ work (char ***message, char **type, long *lines, FILE * id, int tag_table_pos)
 /***************************** keyboard handling ****************************/
       if (key != 0)
 	{
+	  if ((key == keys.print_1) ||
+	      (key == keys.print_2))
+	    {
+	      if (yesno (_ ("Are you sure to print?"), 0) == 1)
+		printnode (message, lines);
+	    }
+/*==========================================================================*/
+	  if ((key == keys.goline_1) ||
+	      (key == keys.goline_2))
+	    {
+	      long newpos;
+	      attrset (bottomline);	/* read user's value */
+	      move (maxy - 1, 0);
+	      echo ();
+	      curs_set (1);
+	      token = getstring (_ ("Enter line: "));
+	      curs_set (0);
+	      noecho ();
+	      move (maxy - 1, 0);
+	      myclrtoeol ();
+	      attrset (normal);
+	      if (token)	/* 
+				 * convert string to long.
+				 * careful with nondigit strings.
+				 */
+		{
+		  int digit_val = 1;
+		  for (i = 0; token[i] != 0; i++)
+		    {
+		      if (!isdigit (token[i]))
+			digit_val = 0;
+		    }
+		  if (digit_val)	/* go to specified line */
+		    {
+		      newpos = atol (token);
+		      newpos -= (maxy - 1);
+		      if ((newpos > 0) && (newpos < Lines - (maxy - 2)))
+			pos = newpos;
+		      else if (newpos > 0)
+			pos = Lines - (maxy - 2);
+		      else
+			pos = 1;
+		    }
+		  xfree (token);
+		  token = 0;
+		}
+	    }
 /*==========================================================================*/
 	  if ((key == keys.shellfeed_1) ||
 	      (key == keys.shellfeed_2))
@@ -105,14 +171,14 @@ work (char ***message, char **type, long *lines, FILE * id, int tag_table_pos)
 	      attrset (bottomline);
 	      move (maxy - 1, 0);
 	      echo ();
-	      leaveok (stdscr, FALSE);
+	      curs_set (1);
 	      token = getstring (_ ("Enter command: "));
 	      noecho ();
 	      move (maxy - 1, 0);
-	      clrtoeol ();
+	      myclrtoeol ();
 	      attrset (normal);
 
-	      endwin ();
+	      myendwin ();
 	      system ("clear");
 	      pipe = popen (token, "w");	/* open pipe */
 	      if (pipe != NULL)
@@ -123,7 +189,7 @@ work (char ***message, char **type, long *lines, FILE * id, int tag_table_pos)
 		  getchar ();
 		}
 	      doupdate ();
-	      leaveok (stdscr, HIDECURSOR);
+	      curs_set (0);
 	      if (pipe == NULL)
 		mvaddstr (maxy - 1, 0, _ ("Operation failed..."));
 	    }
@@ -141,10 +207,10 @@ work (char ***message, char **type, long *lines, FILE * id, int tag_table_pos)
 	  if ((key == keys.refresh_1) ||
 	      (key == keys.refresh_2))
 	    {
-	      endwin ();
+	      myendwin ();
 	      doupdate ();
 	      refresh ();
-	      leaveok (stdscr, HIDECURSOR);
+	      curs_set (0);
 	    }
 /*==========================================================================*/
 	  if ((key == keys.totalsearch_1) ||	/* search in all nodes later than this one */
@@ -154,7 +220,7 @@ work (char ***message, char **type, long *lines, FILE * id, int tag_table_pos)
 	      move (maxy - 1, 0);
 	      attrset (bottomline);
 	      echo ();
-	      leaveok (stdscr, FALSE);
+	      curs_set (1);
 	      if (!searchagain.search)	/* if searchagain key wasn't hit */
 		{
 		  token = getstring (_ ("Enter regexp: "));	/* 
@@ -189,7 +255,7 @@ work (char ***message, char **type, long *lines, FILE * id, int tag_table_pos)
 		  token = xrealloc (token, 2);
 		  strcpy (token, " ");
 		}
-	      leaveok (stdscr, HIDECURSOR);
+	      curs_set (0);
 	      noecho ();
 	      attrset (normal);
 
@@ -234,7 +300,7 @@ work (char ***message, char **type, long *lines, FILE * id, int tag_table_pos)
 
 			fseek (fd, fileoffset, SEEK_SET);
 		      else
-			fseek (fd, tag_table[1].offset, SEEK_SET);
+			fseek (fd, FirstNodeOffset, SEEK_SET);
 		      starttokenpos = ftell (fd);
 
 		      tmp = xmalloc (filelen - starttokenpos + 10);	/* read data */
@@ -243,7 +309,7 @@ work (char ***message, char **type, long *lines, FILE * id, int tag_table_pos)
 
 		      tokenpos = regexp_search (token, tmp);	/* search */
 
-		      if (tokenpos!=-1)	/* if something was found */
+		      if (tokenpos != -1)	/* if something was found */
 			{
 			  tokenpos += starttokenpos;	/* 
 							 * add the offset of
@@ -251,64 +317,73 @@ work (char ***message, char **type, long *lines, FILE * id, int tag_table_pos)
 							 * which wasn't read
 							 * to the memory
 							 */
-			  for (i = TagTableEntries; i >= 1; i--)
+			  {	/* local scope for tmpvar, matched */
+			    int tmpvar = 0, matched = 0;
+			    tag_table[0].offset = 0;
+			    for (i = TagTableEntries; i >= 1; i--)
+			      {
+				if ((tag_table[i].offset > tag_table[tmpvar].offset) &&
+				    ((tag_table[i].offset - indirect[j].offset + FirstNodeOffset) <= tokenpos))
+				  {
+				    return_value = i;
+				    tmpvar = i;
+				    matched = 1;
+				  }
+			      }
+			  }
+			  if (return_value != -1)
+			    /* 
+			     * this means, that indirect entry was found.
+			     */
 			    {
-			      if ((tag_table[i].offset - indirect[j].offset + tag_table[1].offset) <= tokenpos)
-				/* 
-				 * this means, that indirect entry was found.
-				 */
+			      fseek (fd, tag_table[return_value].offset - indirect[j].offset + FirstNodeOffset, SEEK_SET);
+			      /* seek to the found node offset */
+			      while (fgetc (fd) != INFO_TAG);
+			      fgetc (fd);	/* skip newline */
+
+			      aftersearch = 1;
+
+			      while (ftell (fd) < tokenpos)	/* 
+								 * count, how
+								 * many lines
+								 * stands befor
+								 * the token
+								 * line.
+								 */
+
 				{
-				  return_value = i;
-				  fseek (fd, tag_table[i].offset - indirect[j].offset + tag_table[1].offset, SEEK_SET);
-				  /* seek to the found node offset */
-				  while (fgetc (fd) != INFO_TAG);
-				  fgetc (fd);	/* skip newline */
-
-				  aftersearch = 1;
-
-				  while (ftell (fd) < tokenpos)		/* 
-									 * count, how
-									 * many lines
-									 * stands befor
-									 * the token
-									 * line.
-									 */
-
-				    {
-				      int chr = fgetc (fd);
-				      if (chr == '\n')
-					aftersearch++;
-				      else if (chr == EOF)
-					break;
-				    }
-				  if (aftersearch > 1)	/* 
+				  int chr = fgetc (fd);
+				  if (chr == '\n')
+				    aftersearch++;
+				  else if (chr == EOF)
+				    break;
+				}
+			      if (aftersearch > 1)	/* 
 							 * the line where a 
 							 * token is found, is 
 							 * in aftersearch 
 							 */
-				    aftersearch--;
-				  else
-				    aftersearch = 1;
-				  break;
-				}	/* end: if(indirect entry was found) */
-			    }
-			  if (aftersearch)	/* if something was found */
-			    {
-			      if (tmp)	/* free tmp buffer */
-				{
-				  xfree (tmp);
-				  tmp = 0;
-				}
-			      break;
-			    }
-			}	/* end: if(tokenpos) */
-		      if (tmp)	/* free tmp buffer */
-			{
-			  xfree (tmp);
-			  tmp = 0;
+				aftersearch--;
+			      else
+				aftersearch = 1;
+			    }	/* end: if(indirect entry was found) */
 			}
-		      fclose (fd);
-		    }		/* end: search-through-all-indirect-entries loop */
+		      if (aftersearch)	/* if something was found */
+			{
+			  if (tmp)	/* free tmp buffer */
+			    {
+			      xfree (tmp);
+			      tmp = 0;
+			    }
+			  break;
+			}
+		    }		/* end: if(tokenpos) */
+		  if (tmp)	/* free tmp buffer */
+		    {
+		      xfree (tmp);
+		      tmp = 0;
+		    }
+		  fclose (fd);
 		}		/* end: if(indirect) */
 	      else
 		/* if not indirect */
@@ -337,54 +412,62 @@ work (char ***message, char **type, long *lines, FILE * id, int tag_table_pos)
 
 		  tokenpos = regexp_search (token, tmp);	/* search */
 
-		  if (tokenpos!=-1)	/* if we've found something */
+		  if (tokenpos != -1)	/* if we've found something */
 		    {
 		      tokenpos += starttokenpos;	/* 
 							 * add offset of the start
 							 * of search area to this
 							 * token position.
 							 */
-		      for (i = 1; i <= TagTableEntries; i++)
+		      {		/* local scope for tmpvar, matched */
+			int tmpvar = 0, matched = 0;
+			tag_table[0].offset = 0;
+			for (i = TagTableEntries; i >= 1; i--)
+			  {
+			    if ((tag_table[i].offset > tag_table[tmpvar].offset) &&
+				((tag_table[i].offset + FirstNodeOffset) <= tokenpos))
+			      {
+				return_value = i;
+				tmpvar = i;
+				matched = 1;
+			      }
+			  }
+		      }
+		      if (return_value != -1)
+			/* 
+			 * this means, that we've found our entry, and
+			 * we're one position too far with the `i' counter.
+			 */
 			{
-			  if (tag_table[i].offset >= tokenpos)
-			    /* 
-			     * this means, that we've found our entry, and
-			     * we're one position too far with the `i' counter.
-			     */
-			    {
-			      return_value = i - 1;
+			  fseek (id, tag_table[return_value].offset, SEEK_SET);
+			  /* seek to the node, which holds found line */
+			  while (fgetc (id) != INFO_TAG);
+			  fgetc (id);	/* skip newline */
 
-			      fseek (id, tag_table[i - 1].offset, SEEK_SET);
-			      /* seek to the node, which holds found line */
-			      while (fgetc (id) != INFO_TAG);
-			      fgetc (id);	/* skip newline */
-
-			      aftersearch = 1;
-			      while (ftell (id) < tokenpos)	/* 
+			  aftersearch = 1;
+			  while (ftell (id) < tokenpos)		/* 
 								 * count lines
 								 * in found
 								 * node, until
 								 * found line
 								 * is met.
 								 */
-				{
-				  int chr = fgetc (id);
-				  if (chr == '\n')
-				    aftersearch++;
-				  else if (chr == EOF)
-				    break;
-				}
-			      if (aftersearch > 1)
-				aftersearch--;
-			      else
-				aftersearch = 1;
-			      break;
-			      fseek (id, filepos, SEEK_SET);	/* 
+			    {
+			      int chr = fgetc (id);
+			      if (chr == '\n')
+				aftersearch++;
+			      else if (chr == EOF)
+				break;
+			    }
+			  if (aftersearch > 1)
+			    aftersearch--;
+			  else
+			    aftersearch = 1;
+			  fseek (id, filepos, SEEK_SET);	/* 
 								 * seek to old
 								 * filepos.
 								 */
-			    }
-			}	/* end: for loop, counting the aftersearch variable */
+			}
 		    }		/* end: if(tokenpos) <--> token found */
 		  if (tmp)	/* free tmp buffer */
 		    {
@@ -412,7 +495,7 @@ work (char ***message, char **type, long *lines, FILE * id, int tag_table_pos)
 	      move (maxy - 1, 0);
 	      attrset (bottomline);
 	      echo ();
-	      leaveok (stdscr, FALSE);
+	      curs_set (1);
 	      if (!searchagain.search)	/* searchagain handler. see totalsearch */
 		{
 		  token = getstring (_ ("Enter regexp: "));
@@ -430,7 +513,7 @@ work (char ***message, char **type, long *lines, FILE * id, int tag_table_pos)
 		  token = xrealloc (token, 2);
 		  strcpy (token, " ");
 		}
-	      leaveok (stdscr, HIDECURSOR);
+	      curs_set (0);
 	      noecho ();
 	      attrset (normal);
 	      pinfo_re_comp (token);	/* compile the read token */
@@ -449,10 +532,10 @@ work (char ***message, char **type, long *lines, FILE * id, int tag_table_pos)
 		  if (pinfo_re_exec (tmp))	/* execute the search command */
 		    {		/* if found, enter here */
 		      if (pinfo_re_exec (Message[i + 1]))	/* 
-							 * if token was found
-							 * in the second line,
-							 * make pos=i+1.
-							 */
+								   * if token was found
+								   * in the second line,
+								   * make pos=i+1.
+								 */
 			pos = i + 1;
 		      else	/* 
 				   * othwerwise,
@@ -475,6 +558,8 @@ work (char ***message, char **type, long *lines, FILE * id, int tag_table_pos)
 	      xfree (token);	/* free user's search token */
 	      token = 0;
 	      rescan_cursor ();	/* rescan cursor position in the new place */
+	      highlight_searched = 1;
+	      copy_stripped_from_regexp (searchagain.lastsearch, searchword);
 	    }
 /*==========================================================================*/
 	  if ((key == keys.search_again_1) ||	/* search again */
@@ -496,7 +581,9 @@ work (char ***message, char **type, long *lines, FILE * id, int tag_table_pos)
 	      return_value = -1;
 	      move (maxy - 1, 0);
 	      attrset (bottomline);
+	      curs_set (1);
 	      token = getstring (_ ("Enter node name: "));	/* read user's wish */
+	      curs_set (0);
 	      noecho ();
 	      attrset (normal);
 	      for (i = 1; i <= TagTableEntries; i++)
@@ -645,6 +732,13 @@ work (char ***message, char **type, long *lines, FILE * id, int tag_table_pos)
 		  return rval;
 		}
 	    }
+/*==========================================================================*/
+	  if ((key == keys.twoup_1) ||
+	      (key == keys.twoup_2))
+	    {
+	      ungetch (keys.up_1);
+	      ungetch (keys.up_1);
+	    }
 /*==========================================================================*/
 	  if ((key == keys.up_1) ||
 	      (key == keys.up_2))
@@ -694,22 +788,30 @@ work (char ***message, char **type, long *lines, FILE * id, int tag_table_pos)
 	  if ((key == keys.end_1) ||
 	      (key == keys.end_2))
 	    {
-	      pos = Lines - (maxy - 1);
+	      pos = Lines - (maxy - 2);
 	      if (pos < 1)
 		pos = 1;
-	      rescan_cursor ();
+	      cursor = hyperobjectcount - 1;
 	    }
 /*==========================================================================*/
 	  if ((key == keys.pgdn_1) ||
 	      (key == keys.pgdn_2))
 	    {
-	      if (pos + (maxy - 2) < Lines - (maxy - 1))
-		pos += (maxy - 2);
-	      else if (Lines - (maxy - 1) >= 1)
-		pos = Lines - (maxy - 1);
+	      if (pos + (maxy - 2) < Lines - (maxy - 2))
+		{
+		  pos += (maxy - 2);
+		  rescan_cursor ();
+		}
+	      else if (Lines - (maxy - 2) >= 1)
+		{
+		  pos = Lines - (maxy - 2);
+		  cursor = hyperobjectcount - 1;
+		}
 	      else
-		pos = 1;
-	      rescan_cursor ();
+		{
+		  pos = 1;
+		  cursor = hyperobjectcount - 1;
+		}
 	    }
 /*==========================================================================*/
 	  if ((key == keys.pgdn_auto_1) ||
@@ -722,7 +824,7 @@ work (char ***message, char **type, long *lines, FILE * id, int tag_table_pos)
 	  if ((key == keys.home_1) ||
 	      (key == keys.home_2))
 	    {
-	      pos <= 1;
+	      pos = 1;
 	      rescan_cursor ();
 	    }
 /*==========================================================================*/
@@ -742,6 +844,13 @@ work (char ***message, char **type, long *lines, FILE * id, int tag_table_pos)
 	      if (pos == 1)
 		ungetch (keys.upnode_1);
 	    }
+/*==========================================================================*/
+	  if ((key == keys.twodown_1) ||
+	      (key == keys.twodown_2))	/* top+bottom line \|/ */
+	    {
+	      ungetch (keys.down_1);
+	      ungetch (keys.down_1);
+	    }
 /*==========================================================================*/
 	  if ((key == keys.down_1) ||
 	      (key == keys.down_2))	/* top+bottom line \|/ */
@@ -786,8 +895,8 @@ work (char ***message, char **type, long *lines, FILE * id, int tag_table_pos)
 	    {
 	      infohistory.pos[infohistory.length] = pos;
 	      infohistory.cursor[infohistory.length] = cursor;
-	      rval.node = xmalloc (strlen (tag_table[1].nodename) + 1);
-	      strcpy (rval.node, tag_table[1].nodename);
+	      rval.node = xmalloc (strlen (FirstNodeName) + 1);
+	      strcpy (rval.node, FirstNodeName);
 	      rval.file = xmalloc (1);
 	      rval.file[0] = 0;
 	      return rval;
@@ -840,7 +949,7 @@ work (char ***message, char **type, long *lines, FILE * id, int tag_table_pos)
 			    strcpy (tempbuf, httpviewer);
 			    strcat (tempbuf, " ");
 			    strcat (tempbuf, hyperobjects[cursor].node);
-			    endwin ();
+			    myendwin ();
 			    system (tempbuf);
 			    doupdate ();
 			    xfree (tempbuf);
@@ -851,7 +960,7 @@ work (char ***message, char **type, long *lines, FILE * id, int tag_table_pos)
 			    strcpy (tempbuf, ftpviewer);
 			    strcat (tempbuf, " ");
 			    strcat (tempbuf, hyperobjects[cursor].node);
-			    endwin ();
+			    myendwin ();
 			    system (tempbuf);
 			    doupdate ();
 			    xfree (tempbuf);
@@ -862,7 +971,7 @@ work (char ***message, char **type, long *lines, FILE * id, int tag_table_pos)
 			    strcpy (tempbuf, maileditor);
 			    strcat (tempbuf, " ");
 			    strcat (tempbuf, hyperobjects[cursor].node);
-			    endwin ();
+			    myendwin ();
 			    system ("clear");
 			    system (tempbuf);
 			    doupdate ();
@@ -888,9 +997,9 @@ work (char ***message, char **type, long *lines, FILE * id, int tag_table_pos)
 			{
 			  if (hyperobjects[i].line == mouse.y + pos - 1)
 			    {
-			      if (hyperobjects[i].row <= mouse.x - 1)
+			      if (hyperobjects[i].col <= mouse.x - 1)
 				{
-				  if (hyperobjects[i].row + strlen (hyperobjects[i].node) + strlen (hyperobjects[i].file) >= mouse.x - 1)
+				  if (hyperobjects[i].col + strlen (hyperobjects[i].node) + strlen (hyperobjects[i].file) >= mouse.x - 1)
 				    {
 				      if (hyperobjects[i].type < HIGHLIGHT)
 					{
@@ -907,9 +1016,9 @@ work (char ***message, char **type, long *lines, FILE * id, int tag_table_pos)
 			  {
 			    if (hyperobjects[i].line == mouse.y + pos - 1)
 			      {
-				if (hyperobjects[i].row <= mouse.x - 1)
+				if (hyperobjects[i].col <= mouse.x - 1)
 				  {
-				    if (hyperobjects[i].row + strlen (hyperobjects[i].node) + strlen (hyperobjects[i].file) >= mouse.x - 1)
+				    if (hyperobjects[i].col + strlen (hyperobjects[i].node) + strlen (hyperobjects[i].file) >= mouse.x - 1)
 				      {
 					if (hyperobjects[i].type < HIGHLIGHT)
 					  {
@@ -935,9 +1044,9 @@ work (char ***message, char **type, long *lines, FILE * id, int tag_table_pos)
 			{
 			  if (hyperobjects[i].line == mouse.y + pos - 1)
 			    {
-			      if (hyperobjects[i].row <= mouse.x - 1)
+			      if (hyperobjects[i].col <= mouse.x - 1)
 				{
-				  if (hyperobjects[i].row + strlen (hyperobjects[i].node) + strlen (hyperobjects[i].file) >= mouse.x - 1)
+				  if (hyperobjects[i].col + strlen (hyperobjects[i].node) + strlen (hyperobjects[i].file) >= mouse.x - 1)
 				    {
 				      if (hyperobjects[i].type < HIGHLIGHT)
 					{
@@ -954,9 +1063,9 @@ work (char ***message, char **type, long *lines, FILE * id, int tag_table_pos)
 			  {
 			    if (hyperobjects[i].line == mouse.y + pos - 1)
 			      {
-				if (hyperobjects[i].row <= mouse.x - 1)
+				if (hyperobjects[i].col <= mouse.x - 1)
 				  {
-				    if (hyperobjects[i].row + strlen (hyperobjects[i].node) + strlen (hyperobjects[i].file) >= mouse.x - 1)
+				    if (hyperobjects[i].col + strlen (hyperobjects[i].node) + strlen (hyperobjects[i].file) >= mouse.x - 1)
 				      {
 					if (hyperobjects[i].type < HIGHLIGHT)
 					  {
@@ -980,8 +1089,16 @@ work (char ***message, char **type, long *lines, FILE * id, int tag_table_pos)
 #endif
 /*****************************************************************************/
 	}
-      if (key == keys.quit_2)
-	break;
+      if ((key == keys.quit_2) || (key == keys.quit_1))
+	{
+	  if (!ConfirmQuit)
+	    break;
+	  else
+	    {
+	      if (yesno (_ ("Are you sure to quit?"), QuitConfirmDefault))
+		break;
+	    }
+	}
     }
   return rval;
 }
@@ -1013,7 +1130,7 @@ getnodeoffset (int tag_table_pos, int *Indstart)	/* count node offset in file */
 	{
 	  if (indirect[i].offset <= tag_table[tag_table_pos].offset)
 	    {
-	      fileoffset += (tag_table[tag_table_pos].offset - indirect[i].offset + tag_table[1].offset);
+	      fileoffset += (tag_table[tag_table_pos].offset - indirect[i].offset + FirstNodeOffset);
 	      indirectstart = i;
 	      break;
 	    }
