@@ -1,5 +1,5 @@
 #include "common_includes.h"
-#include <regex.h>
+#include <ctype.h>
 #include <sys/stat.h>
 
 #define HTTPSECTION 100
@@ -10,13 +10,22 @@ void loadmanual (FILE * id);	/* load manual */
 int manualwork ();		/* handle keyboard */
 void rescan_selected ();	/* scan for potential link to select on
 				   viewed manual page */
-void showmanualscreen ();	/* self explainig */
+void showmanualscreen ();	/* self explanatory */
 void mvaddstr_manual (int y, int x, char *str);		/* mvaddstr with bold/italic */
 void add_highlights ();		/* adds highlights to a painted screen */
 void strip_manual (char *buf);	/* strips line from formatting characters */
-void man_initializelinks (char *line);	/* initialize links in a line */
+void man_initializelinks (char *line, int carry);
+ /* 
+  * Initialize links in a line .
+  * Links are entries of form
+  * reference (section), and are stored
+  * in `manuallinks' var, described 
+  * bellow.
+  */
 int is_in_manlinks (char *in, char *find);
 
+void printmanual (char **Message, long Lines);
+
 char **manual = 0;		/* line by line stored manual */
 int ManualLines = 0;		/* number of lines in manual */
 int selected = -1;		/* number of selected link (offset in 'manuallinks',
@@ -31,21 +40,28 @@ typedef struct
     int selected;		/* what was last selected on this page */
     int pos;			/* what was the last manualpos */
   }
-manhistory;
+manhistory;			/* 
+				 * type for the `lastread' history entries, when viewing
+				 * man pages.
+				 */
 
-manhistory *manualhistory = 0;
+manhistory *manualhistory = 0;	/* manual lastread history */
 int manualhistorylength = 0;	/* length of the above table - 1 */
 
 typedef struct
   {				/* struct for hypertext references */
     int line;			/* line of the manpage, where the reference
 				   is */
-    int row;			/* row of that line */
+    int col;			/* column of that line */
     char *name;			/* name of the reference */
     char section[32];		/* section of the reference */
     int section_mark;
+    int carry;			/* determine whether there is a hyphen above */
   }
-manuallink;
+manuallink;			/*
+				 * this structure describes a hyperlink
+				 * in manual viewer
+				 */
 
 manuallink *manuallinks = 0;	/* a set of manual references of man page */
 
@@ -56,12 +72,12 @@ int historical = 0;		/* semaphore for checking if it's a history (left
 
 
 void
-manual_free_buffers ()
+manual_free_buffers ()		/* free buffers allocated by current man page */
 {
   int i;
   if (manual)			/* first free previously allocated memory */
     {				/* for the manual itself... */
-      for (i = 0; i < ManualLines; i++)
+      for (i = 0; i <= ManualLines; i++)
 	{
 	  xfree (manual[i]);
 	}
@@ -83,34 +99,231 @@ manual_free_buffers ()
 }
 
 void
-set_initial_history (char *name)
+set_initial_history (char *name)	/* 
+					 * initialize history variables for 
+					 * manual pages.
+					 */
 {
   manualhistory = xmalloc (sizeof (manhistory));	/* one object of
 							   array */
   strcpy (manualhistory[0].name, name);		/* filename->name */
-  strcpy(manualhistory[0].sect,"");	/* section unknown */
+  strcpy (manualhistory[0].sect, "");	/* section unknown */
   manualhistory[0].selected = -1;	/* selected unknown */
   manualhistory[0].pos = 0;	/* pos=0 */
 }
 
+void
+construct_manualname (char *buf, int which)	/* construct man name; take
+						   care about carry */
+{
+  if (!manuallinks[which].carry)
+    {
+      strcpy (buf, manuallinks[which].name);
+      return;
+    }
+  else
+    {
+      char *base = xmalloc (1024);
+      char *ptr;
+      int tmppos;
+      strcpy (base, manual[manuallinks[which].line - 1]);
+      strip_manual (base);
+      ptr = base + strlen (base) - 3;
+      while (((isalpha (*ptr)) || (*ptr == '.')) && (ptr > base))
+	ptr--;
+      strcpy (buf, ptr);
+      tmppos = strlen (buf);
+      if (tmppos > 1);
+      buf[tmppos - 2] = 0;
+      strcat (buf, manuallinks[which].name);
+      xfree (base);
+    }
+}
+
 int
-handlemanual (char *name)
+handlemanual (char *name)	/*
+				 * this is something like main() function
+				 * for the manual viewer code.
+				 */
 {
   int return_value = 0;
   struct stat statbuf;
+  FILE *id, *source;
+
+  char **ignored_entries;
+  char manualname[256];
   char cmd[256];
-  FILE *id;
+  char location[256];
+  char line[1025];
+  char *raw_tempfilename = 0;
+  char *end, *prev;
+  size_t macroline_size;
+  int ignored_items = 0, i = 0;
+  char zipped = 0;
+
   if (tmpfilename1)
     {
       unlink (tmpfilename1);
       xfree (tmpfilename1);
     }
   tmpfilename1 = tempnam ("/tmp", NULL);
-  snprintf (cmd, 255, "man %s %s %s > %s", 
-            ManOptions, 
-            name, 
-            StderrRedirection,
-            tmpfilename1);
+
+#ifdef getmaxyx
+  init_curses ();
+  getmaxyx (stdscr, maxy, maxx);	/* if ncurses, get maxx and maxy */
+  myendwin ();
+  {
+    char tmp[4];		/* set MANWIDTH environment variable */
+    snprintf (tmp, 4, "%d", maxx);
+    setenv ("MANWIDTH", tmp, 1);
+  }
+#else
+  maxx = 80;
+  maxy = 25;			/* otherwise hardcode 80x25... */
+#endif /* getmaxyx */
+
+/****************************************************************************
+ *                    Ignore macros part: BEGIN                             *
+ * PS: Siewca: I still expect that you'll isolate it to a single procedure  *
+ ****************************************************************************/
+  if (ignoredmacros)		/* if the pointer is non-null */
+    if (*ignoredmacros && strlen (ignoredmacros))	/* if there are some macros */
+      {				/* that should be ignored   */
+	*location = '\0';
+	snprintf (cmd, 255, "man -W %s %s",	/* we need to know the path */
+		  ManOptions,
+		  name);
+	id = popen (cmd, "r");
+	if (!id)
+	  {
+	    printf (_ ("Error: Cannot call man command.\n"));
+	    return 1;
+	  }
+	fflush (id);
+	fgets (location, 255, id);
+	pclose (id);
+
+	if (*location == '\0')
+	  {
+	    printf (_ ("Error: No manual page found either.\n"));
+	    if (use_apropos)
+	      {
+		printf (_ ("Appropriate pages:\n"));
+		snprintf (cmd, 255, "apropos %s|cat %s", name, StderrRedirection);
+		system (cmd);
+	      }
+	    return 1;
+	  }
+
+
+	ignored_items++;
+	prev = ignoredmacros;
+	while ((end = strchr (prev, ':')))	/* counting items */
+	  {
+	    ignored_items++;
+	    prev = end + 1;
+	  }
+
+	ignored_entries = (char **) xmalloc (ignored_items * sizeof (char **));
+	ignored_entries[0] = ignoredmacros;
+	prev = ignoredmacros;
+	i = 0;
+	while ((end = strchr (prev, ':')))	/* creating pointers */
+	  {
+	    *end = '\0';
+	    prev = end + 1;
+	    i++;
+	    ignored_entries[i] = prev;
+	  }
+
+	if ((prev = rindex (location, '\n')))
+	  *prev = '\0';		/* removing newline */
+
+	prev = index (location, '\0');	/* checking if it's comressed */
+	if ((strlen (location)) > 3
+	    && ((*(prev - 1) == 'Z' && *(prev - 2) == '.')
+	 || (*(prev - 1) == 'z' && *(prev - 2) == 'g' && *(prev - 3) == '.')
+	    )
+	  )
+	  {
+	    printf ("%s %s\n", _ ("Calling gunzip for"), location);
+	    snprintf (cmd, 255, "gunzip -c %s", location);
+	    source = popen (cmd, "r");
+	    zipped = 1;
+	    if (!source)
+	      {
+		printf (_ ("Couldn't call gunzip.\n"));
+		return 1;
+	      }
+	  }
+	else
+	  source = fopen (location, "r");	/* from cmd output  */
+	name = tempnam ("/tmp", NULL);
+	raw_tempfilename = name;
+	id = fopen (name, "w");
+
+	while (!feof (source))	/* we read until eof */
+	  {
+	    if (fgets (line, 1024, source) == NULL)
+	      line[0] = '\0';
+
+	    if (line[0] != '.' || (strlen (line)) < (size_t) 2)		/* macro starts */
+	      {			/* with da DOT  */
+		fprintf (id, "%s", line);
+		continue;
+	      }
+	    else
+	      while (i >= 0)
+		{
+		  macroline_size = strlen (ignored_entries[i]);
+		  if (strlen (line + 1) < macroline_size)
+		    macroline_size = strlen (line + 1);
+		  if ((strncmp (ignored_entries[i], line + 1, macroline_size)) == 0
+		      && (*(line + 1 + (int) macroline_size) == ' '
+			  || *(line + 1 + (int) macroline_size) == '\n'
+			  || *(line + 1 + (int) macroline_size) == '\t'))
+		    {
+		      if (quote_ignored)
+			{
+			  if ((prev = rindex (line, '\n')))
+			    *prev = '\0';
+			  sprintf (cmd, "\n.br\n.nf\n[ [pinfo] - %s: %.42s", _ ("IGNORING"), line);
+			  if ((strlen (line)) > (size_t) 42)
+			    strcat (cmd, " (...)]\n.fi\n");
+			  else
+			    strcat (cmd, " ]\n.fi\n");
+			}
+		      else
+			{
+			  sprintf (cmd, ".\\\" removed macro: %.42s", line);
+			  if ((strlen (line)) > (size_t) 42)
+			    strcat (cmd, " (...)");
+			}
+		      strcpy (line, cmd);
+		      break;
+		    }
+		  i--;
+		}
+
+	    fprintf (id, "%s", line);
+	    i = ignored_items - 1;
+	  }			/* while (!feof (source)) */
+	if (zipped)
+	  pclose (source);
+	else
+	  fclose (source);
+	fclose (id);
+	free (ignored_entries);
+      }				/* if (ignored_macros... */
+/****************************************************************************
+ *                    Ignore macros part: END                               *
+ ****************************************************************************/
+
+  snprintf (cmd, 255, "man %s %s %s > %s",
+	    ManOptions,
+	    name,
+	    StderrRedirection,
+	    tmpfilename1);
   if (system (cmd) != 0)
     {
       printf (_ ("Error: No manual page found either.\n"));
@@ -124,6 +337,8 @@ handlemanual (char *name)
     }
   id = fopen (tmpfilename1, "r");
   init_curses ();
+
+
   loadmanual (id);		/* load manual to memory */
   fclose (id);
   set_initial_history (name);
@@ -143,12 +358,15 @@ handlemanual (char *name)
 	  if (return_value != -2)	/* key_back is not pressed;
 					   and return_value is an
 					   offset to manuallinks */
-	    snprintf (cmd, 255, "man %s %s %s %s > %s",
-	    	      ManOptions,
-		      manuallinks[return_value].section,
-		      manuallinks[return_value].name,
-		      StderrRedirection,
-		      tmpfilename2);
+	    {
+	      construct_manualname (manualname, return_value);
+	      snprintf (cmd, 255, "man %s %s %s %s > %s",
+			ManOptions,
+			manuallinks[return_value].section,
+			manualname,
+			StderrRedirection,
+			tmpfilename2);
+	    }
 	  else
 	    /* key_back was pressed */
 	    {
@@ -188,8 +406,8 @@ handlemanual (char *name)
 		    {
 		      manualhistorylength++;
 		      manualhistory = xrealloc (manualhistory, (manualhistorylength + 2) * sizeof (manhistory));
-		      strcpy (manualhistory[manualhistorylength].name, manuallinks[return_value].name);
-		      strcpy(manualhistory[manualhistorylength].sect,manuallinks[return_value].section);
+		      strcpy (manualhistory[manualhistorylength].name, manualname);	/* we can write so since this code applies only when it's not a history call */
+		      strcpy (manualhistory[manualhistorylength].sect, manuallinks[return_value].section);
 		    }
 		  loadmanual (id);	/* loading manual page and its defaults... */
 		  fclose (id);
@@ -207,26 +425,27 @@ handlemanual (char *name)
 	}
     }
   while (return_value != -1);
-  return 0;
+  if (raw_tempfilename)
+    unlink (raw_tempfilename);	/* we were using temporary */
+  return 0;			/* raw-manpage for scaning */
 }
 
 void
-loadmanual (FILE * id)
+loadmanual (FILE * id)		/* loads manual from given filedescriptor */
 {
-  int i;
   char prevlinechar = 0;
   int cutheader = 0;		/* tmp variable, set after reading first
 				   nonempty line of input */
+  int carryflag = 0;
   manualpos = 0;
   manual_free_buffers ();
   manual = xmalloc (sizeof (char *));
   manuallinks = xmalloc (sizeof (manuallinks));
   manual[ManualLines] = xmalloc (1024);
+
   while (!feof (id))		/* we read until eof */
     {
       char *tmp;
-      char *link;
-      int tmpcnt = 0;
       if (fgets (manual[ManualLines], 1024, id) == NULL)	/* it happens 
 								   sometimes, 
 								   that the last 
@@ -239,6 +458,8 @@ loadmanual (FILE * id)
 					   buffer, what 
 					   confuses 
 					   strlen */
+
+
       if (cutheader)
 	{
 	  if (strcmp (manual[cutheader], manual[ManualLines]) == 0)
@@ -247,6 +468,12 @@ loadmanual (FILE * id)
 	      manual[ManualLines][1] = 0;
 	    }
 	}
+      if (FilterB7)
+	{
+	  char *filter_pos = index (manual[ManualLines], 0xb7);
+	  if (filter_pos)
+	    *filter_pos = 'o';
+	}
       if (CutManHeaders)
 	if (!cutheader)
 	  {
@@ -262,37 +489,43 @@ loadmanual (FILE * id)
 	}
       else
 	{
+	  int manlinelen = strlen (manual[ManualLines]);
 	  manual[ManualLines] = xrealloc (manual[ManualLines],
-					  strlen (manual[ManualLines]) + 10);
+					  manlinelen + 10);
 
-	  tmp = xmalloc (strlen (manual[ManualLines]) + 10);	/* temporary 
-								   variable for 
-								   determining 
-								   hypertextuality
-								   of fields */
+	  tmp = xmalloc (manlinelen + 10);	/* temporary 
+						   variable for 
+						   determining 
+						   hypertextuality
+						   of fields */
 
 	  strcpy (tmp, manual[ManualLines]);
 
 	  strip_manual (tmp);	/* remove formatting chars */
-	  man_initializelinks (tmp);
+	  man_initializelinks (tmp, carryflag);
+	  carryflag = 0;
+	  if (manlinelen > 1)
+	    if (manual[ManualLines][manlinelen - 2] == '-')
+	      carryflag = 1;
 	  xfree (tmp);		/* free temporary buffer */
 	  prevlinechar = manual[ManualLines][0];
 	  ManualLines++;	/* increase the number of man lines */
-	  manual = xrealloc (manual, (ManualLines + 1) * sizeof (char *));
+	  manual = xrealloc (manual, (ManualLines + 5) * sizeof (char *));
 	  /* and realloc manual to add an empty
 	     space for next entry of manual line */
 	  manual[ManualLines] = xmalloc (1024);
 	}
     }
+
 }
 
 void
-man_initializelinks (char *tmp)
+man_initializelinks (char *tmp, int carry)	/* initializes hyperlinks in manual */
 {
   int tmpcnt = strlen (tmp) + 1;	/* set tmpcnt to the trailing zero of tmp */
   char *link = tmp;
   char *urlstart, *urlend;
-  int i,b;
+  int i, b;
 /******************************************************************************
  * handle url refrences                                                       *
  *****************************************************************************/
@@ -301,9 +534,9 @@ man_initializelinks (char *tmp)
       urlend = findurlend (urlstart);	/* always successfull */
       manuallinks = xrealloc (manuallinks, sizeof (manuallink) * (ManualLinks + 3));
       manuallinks[ManualLinks].line = ManualLines;
-      manuallinks[ManualLinks].row = urlstart - tmp;
-      strcpy(manuallinks[ManualLinks].section,"HTTPSECTION");
-      manuallinks[ManualLinks].section_mark=HTTPSECTION;
+      manuallinks[ManualLinks].col = urlstart - tmp;
+      strcpy (manuallinks[ManualLinks].section, "HTTPSECTION");
+      manuallinks[ManualLinks].section_mark = HTTPSECTION;
       manuallinks[ManualLinks].name = xmalloc (urlend - urlstart + 10);
       strncpy (manuallinks[ManualLinks].name, urlstart, urlend - urlstart);
       manuallinks[ManualLinks].name[urlend - urlstart] = 0;
@@ -314,9 +547,9 @@ man_initializelinks (char *tmp)
       urlend = findurlend (urlstart);	/* always successfull */
       manuallinks = xrealloc (manuallinks, sizeof (manuallink) * (ManualLinks + 3));
       manuallinks[ManualLinks].line = ManualLines;
-      manuallinks[ManualLinks].row = urlstart - tmp;
-      strcpy(manuallinks[ManualLinks].section,"FTPSECTION");
-      manuallinks[ManualLinks].section_mark=FTPSECTION;
+      manuallinks[ManualLinks].col = urlstart - tmp;
+      strcpy (manuallinks[ManualLinks].section, "FTPSECTION");
+      manuallinks[ManualLinks].section_mark = FTPSECTION;
       manuallinks[ManualLinks].name = xmalloc (urlend - urlstart + 10);
       strncpy (manuallinks[ManualLinks].name, urlstart, urlend - urlstart);
       manuallinks[ManualLinks].name[urlend - urlstart] = 0;
@@ -327,9 +560,9 @@ man_initializelinks (char *tmp)
       urlend = findurlend (urlstart);	/* always successfull */
       manuallinks = xrealloc (manuallinks, sizeof (manuallink) * (ManualLinks + 3));
       manuallinks[ManualLinks].line = ManualLines;
-      manuallinks[ManualLinks].row = urlstart - tmp;
-      strcpy(manuallinks[ManualLinks].section, "MAILSECTION");
-      manuallinks[ManualLinks].section_mark=MAILSECTION;
+      manuallinks[ManualLinks].col = urlstart - tmp;
+      strcpy (manuallinks[ManualLinks].section, "MAILSECTION");
+      manuallinks[ManualLinks].section_mark = MAILSECTION;
       manuallinks[ManualLinks].name = xmalloc (urlend - urlstart + 10);
       strncpy (manuallinks[ManualLinks].name, urlstart, urlend - urlstart);
       manuallinks[ManualLinks].name[urlend - urlstart] = 0;
@@ -348,7 +581,7 @@ man_initializelinks (char *tmp)
 	  if ((temp = strchr (link, ')')))	/* look for the closing bracket */
 	    {
 	      char *p_t1, *p_t;
-	      p_t = p_t1 = xmalloc ((strlen (link) + 1) * sizeof (char));
+	      p_t = p_t1 = xmalloc ((strlen (link) + 10) * sizeof (char));
 	      for (++link; link != temp; *p_t++ = *link++);
 	      *p_t = '\0';
 	      link -= (strlen (p_t1) + sizeof (char));
@@ -380,22 +613,35 @@ man_initializelinks (char *tmp)
 		  /* now we have needed string in i..breakpos. We need now to realloc the manuallinks table to make free space for new entry */
 		  manuallinks = xrealloc (manuallinks, sizeof (manuallink) * (ManualLinks + 3));
 		  manuallinks[ManualLinks].line = ManualLines;
-		  manuallinks[ManualLinks].row = i;
-		  if(LongManualLinks)
+		  manuallinks[ManualLinks].col = i;
+		  if (LongManualLinks)
 		    {
-  		      for(b=1;link[b]!=')';b++)
-  		        manuallinks[ManualLinks].section[b-1] = tolower(link[b]);
-  		      manuallinks[ManualLinks].section[b-1]=0;
-  		    }
-  		  else
-  		    {
-  		      manuallinks[ManualLinks].section[0]=link[1];
-  		      manuallinks[ManualLinks].section[1]=0;
-  		    }
-  		  manuallinks[ManualLinks].section_mark=0;
+		      for (b = 1; link[b] != ')'; b++)
+			manuallinks[ManualLinks].section[b - 1] = tolower (link[b]);
+		      manuallinks[ManualLinks].section[b - 1] = 0;
+		    }
+		  else
+		    {
+		      manuallinks[ManualLinks].section[0] = link[1];
+		      manuallinks[ManualLinks].section[1] = 0;
+		    }
+		  manuallinks[ManualLinks].section_mark = 0;
 		  manuallinks[ManualLinks].name = xmalloc ((breakpos - i) + 10);
 		  strcpy (manuallinks[ManualLinks].name, tmp + i);
 		  tmp[breakpos] = tempchar;
+		  for (b = i - 1; b >= 0; b--)	/* check whether this is a carry'ed
+						 * entry (i.e. in the previous line
+						 * there was `-' at end, and this is
+						 * the first word of this line */
+		    {
+		      if (b > 0)
+			if ((tmp[b] != ' ') && (tmp[b] != '\t'))
+			  break;
+		    }
+		  if (b >= 0)
+		    manuallinks[ManualLinks].carry = 0;
+		  else
+		    manuallinks[ManualLinks].carry = carry;
 		  ManualLinks++;	/* increase the number of entries */
 		}
 	      xfree ((void *) p_t1);
@@ -413,15 +659,22 @@ man_initializelinks (char *tmp)
 }
 
 int
-manualwork ()
+manualwork ()			/* viewer function. Handles keyboard actions--main event loop */
 {
   FILE *pipe;			/* for user's shell commands */
   char *token;			/* a temporary buffer */
   char *tmp;			/* again the same */
   int key = 0;			/* key, which contains the value entered by user */
   int i, selectedchanged;	/* tmp values */
+  int highlight_searched = 0;	/* is it time to highlight after search? */
+  char searchword[256];		/* buffer for searchword highlighter */
 #ifdef getmaxyx
   getmaxyx (stdscr, maxy, maxx);	/* if ncurses, get maxx and maxy */
+  {
+    char tmp[4];		/* set MANWIDTH environment variable */
+    snprintf (tmp, 4, "%d", maxx);
+    setenv ("MANWIDTH", tmp, 1);
+  }
 #else
   maxx = 80;
   maxy = 25;			/* otherwise hardcode 80x25... */
@@ -440,14 +693,33 @@ manualwork ()
 				   selected on currently
 				   viewed page */
   erase ();			/* clean screen */
-  while (key != keys.quit_1)	/* user events loop. finish when
+  while (1)			/* user events loop. finish when
 				   key_quit */
     {
       nodelay (stdscr, TRUE);	/* make getch not wait for user */
-      key = pinfo_getch ();		/* action -- return ERR */
+      key = pinfo_getch ();	/* action -- return ERR */
       if (key == ERR)		/* if there was nothing in buffer */
 	{
 	  showmanualscreen ();	/* then show screen */
+	  if (highlight_searched)
+	    {
+	      char *tmp = xmalloc (strlen (manual[manualpos]) + 10);
+	      char *highlightstart = 0;
+	      strcpy (tmp, manual[manualpos]);
+	      strip_manual (tmp);	/* we must do it without counting nonprintable chars */
+	      highlightstart = strstr (tmp, searchword);
+	      attrset (searchhighlight);
+	      if (highlightstart)
+		{
+		  mvaddstr (1,
+			    calculate_len (tmp, highlightstart),
+			    searchword);
+		}
+	      highlight_searched = 0;
+	      attrset (normal);
+	      move (0, 0);
+	      xfree (tmp);
+	    }
 	  nodelay (stdscr, FALSE);
 	  key = pinfo_getch ();
 	}
@@ -457,22 +729,94 @@ manualwork ()
 /************************ keyboard handling **********************************/
       if (key != 0)
 	{
+	  if ((key == keys.print_1) ||
+	      (key == keys.print_2))
+	    {
+	      if (yesno (_ ("Are you sure to print?"), 0))
+		printmanual (manual, ManualLines);
+	    }
+/*==========================================================================*/
+	  if ((key == keys.goto_1) ||
+	      (key == keys.goto_2))
+	    {
+	      manuallinks = xrealloc (manuallinks, (ManualLinks + 1) * (sizeof (manuallink) + 3));
+
+	      attrset (bottomline);	/* get user's value */
+	      move (maxy - 1, 0);
+	      echo ();
+	      curs_set (1);
+	      manuallinks[ManualLinks].name = getstring (_ ("Enter manual name: "));
+	      curs_set (0);
+	      noecho ();
+	      move (maxy - 1, 0);
+	      myclrtoeol ();
+	      attrset (normal);
+
+	      manuallinks[ManualLinks].carry = 0;
+	      manuallinks[ManualLinks].section_mark = 0;
+	      strcpy (manuallinks[ManualLinks].section, " ");
+	      manuallinks[ManualLinks].line = -1;
+	      manuallinks[ManualLinks].col = -1;
+	      ManualLinks++;
+	      return ManualLinks - 1;
+	    }
+/*==========================================================================*/
+	  if ((key == keys.goline_1) ||
+	      (key == keys.goline_2))
+	    {
+	      long newpos;
+	      attrset (bottomline);	/* get user's value */
+	      move (maxy - 1, 0);
+	      echo ();
+	      curs_set (1);
+	      token = getstring (_ ("Enter line: "));
+	      curs_set (0);
+	      noecho ();
+	      move (maxy - 1, 0);
+	      myclrtoeol ();
+	      attrset (normal);
+	      if (token)	/*
+				 * convert string to long.
+				 * careful with nondigit strings.
+				 */
+		{
+		  int digit_val = 1;
+		  for (i = 0; token[i] != 0; i++)
+		    {
+		      if (!isdigit (token[i]))
+			digit_val = 0;
+		    }
+		  if (digit_val)	/* move cursor position */
+		    {
+		      newpos = atol (token);
+		      newpos -= (maxy - 1);
+		      if ((newpos >= 0) && (newpos < ManualLines - (maxy - 2)))
+			manualpos = newpos;
+		      else if (newpos > 0)
+			manualpos = ManualLines - (maxy - 2);
+		      else
+			manualpos = 0;
+		    }
+		  xfree (token);
+		  token = 0;
+		}
+	    }
 /*===========================================================================*/
 	  if ((key == keys.shellfeed_1) ||
 	      (key == keys.shellfeed_2))
 	    {
 	      /* get command name */
-	      leaveok (stdscr, FALSE);
+	      curs_set (1);
 	      attrset (bottomline);
 	      move (maxy - 1, 0);
 	      echo ();
 	      token = getstring (_ ("Enter command: "));	/* get users cmd */
 	      noecho ();
 	      move (maxy - 1, 0);
-	      clrtoeol ();
+	      myclrtoeol ();
 	      attrset (normal);
 
-	      endwin ();
+	      myendwin ();
 	      system ("clear");
 	      pipe = popen (token, "w");	/* open pipe */
 	      if (pipe != NULL)
@@ -483,16 +827,16 @@ manualwork ()
 		}
 	      getchar ();
 	      doupdate ();
-	      leaveok (stdscr, HIDECURSOR);
+	      curs_set (0);
 	    }
 /*===========================================================================*/
 	  if ((key == keys.refresh_1) ||
 	      (key == keys.refresh_2))
 	    {
-	      endwin ();
+	      myendwin ();
 	      doupdate ();
 	      refresh ();
-	      leaveok (stdscr, HIDECURSOR);
+	      curs_set (0);
 	    }
 /*===========================================================================*/
 	  if ((key == keys.search_1) ||		/* search in current node */
@@ -501,7 +845,7 @@ manualwork ()
 	      move (maxy - 1, 0);	/* procedure of getting regexp string */
 	      attrset (bottomline);
 	      echo ();
-	      leaveok (stdscr, FALSE);
+	      curs_set (1);
 	      if (!searchagain.search)	/* 
 					 * searchagain handler. see
 					 * keys.totalsearch at mainfunction.c
@@ -523,10 +867,10 @@ manualwork ()
 		  token = xrealloc (token, 2);
 		  strcpy (token, " ");
 		}
-	      leaveok (stdscr, HIDECURSOR);
+	      curs_set (0);
 	      noecho ();
 	      move (maxy - 1, 0);
-	      clrtoeol ();
+	      myclrtoeol ();
 	      attrset (normal);
 	      pinfo_re_comp (token);	/* compile regexp expression */
 	      for (i = manualpos + 1; i < ManualLines - 1; i++)
@@ -560,6 +904,8 @@ manualwork ()
 		}
 	      xfree (token);
 	      rescan_selected ();
+	      highlight_searched = 1;
+	      copy_stripped_from_regexp (searchagain.lastsearch, searchword);
 	    }
 /*===========================================================================*/
 	  if ((key == keys.search_again_1) ||	/* search again */
@@ -571,6 +917,13 @@ manualwork ()
 		  ungetch (searchagain.type);
 		}
 	    }
+/*===========================================================================*/
+	  if ((key == keys.twoup_1) ||
+	      (key == keys.twoup_2))
+	    {
+	      ungetch (keys.up_1);
+	      ungetch (keys.up_1);
+	    }
 /*===========================================================================*/
 	  if ((key == keys.up_1) ||
 	      (key == keys.up_2))
@@ -616,19 +969,27 @@ manualwork ()
 	      (key == keys.end_2))
 	    {
 	      manualpos = ManualLines - (maxy - 1);
-	      rescan_selected ();
+	      selected = ManualLinks - 1;
 	    }
 /*===========================================================================*/
 	  if ((key == keys.pgdn_1) ||
 	      (key == keys.pgdn_2))
 	    {
 	      if (manualpos + (maxy - 2) < ManualLines - (maxy - 1))
-		manualpos += (maxy - 2);
+		{
+		  manualpos += (maxy - 2);
+		  rescan_selected ();
+		}
 	      else if (ManualLines - (maxy - 1) >= 1)
-		manualpos = ManualLines - (maxy - 1);
+		{
+		  manualpos = ManualLines - (maxy - 1);
+		  selected = ManualLinks - 1;
+		}
 	      else
-		manualpos = 0;
-	      rescan_selected ();
+		{
+		  manualpos = 0;
+		  selected = ManualLinks - 1;
+		}
 	    }
 /*===========================================================================*/
 	  if ((key == keys.home_1) ||
@@ -647,6 +1008,13 @@ manualwork ()
 		manualpos = 1;
 	      rescan_selected ();
 	    }
+/*===========================================================================*/
+	  if ((key == keys.twodown_1) ||
+	      (key == keys.twodown_2))	/* top+bottom line \|/ */
+	    {			/* see keys.up for comments */
+	      ungetch (keys.down_1);
+	      ungetch (keys.down_1);
+	    }
 /*===========================================================================*/
 	  if ((key == keys.down_1) ||
 	      (key == keys.down_2))	/* top+bottom line \|/ */
@@ -693,48 +1061,49 @@ manualwork ()
 	    {
 	      manualhistory[manualhistorylength].pos = manualpos;
 	      manualhistory[manualhistorylength].selected = selected;
-	      if ((manuallinks[selected].line >= manualpos) &&
-		  (manuallinks[selected].line < manualpos + (maxy - 1)))
-		{
-		      if (!strncmp(manuallinks[selected].section,"HTTPSECTION",11))
-			{
-			  char *tempbuf = xmalloc (strlen (manuallinks[selected].name) + strlen (httpviewer) + 10);
-			  strcpy (tempbuf, httpviewer);
-			  strcat (tempbuf, " ");
-			  strcat (tempbuf, manuallinks[selected].name);
-			  endwin ();
-			  system (tempbuf);
-			  doupdate ();
-			  xfree (tempbuf);
-			}
-		      else if (!strncmp(manuallinks[selected].section,"FTPSECTION",10))
-			{
-			  char *tempbuf = xmalloc (strlen (manuallinks[selected].name) + strlen (ftpviewer) + 10);
-			  strcpy (tempbuf, httpviewer);
-			  strcat (tempbuf, " ");
-			  strcat (tempbuf, manuallinks[selected].name);
-			  endwin ();
-			  system (tempbuf);
-			  doupdate ();
-			  xfree (tempbuf);
-			}
-		      else if (!strncmp(manuallinks[selected].section,"MAILSECTION",11))
-			{
-			  char *tempbuf = xmalloc (strlen (manuallinks[selected].name) + strlen (maileditor) + 10);
-			  strcpy (tempbuf, maileditor);
-			  strcat (tempbuf, " ");
-			  strcat (tempbuf, manuallinks[selected].name);
-			  endwin ();
-			  system ("clear");
-			  system (tempbuf);
-			  doupdate ();
-			  xfree (tempbuf);
-			}
-		      else
-		        {
-		          return selected;
-		        }
-		}
+	      if (selected >= 0)
+		if ((manuallinks[selected].line >= manualpos) &&
+		    (manuallinks[selected].line < manualpos + (maxy - 1)))
+		  {
+		    if (!strncmp (manuallinks[selected].section, "HTTPSECTION", 11))
+		      {
+			char *tempbuf = xmalloc (strlen (manuallinks[selected].name) + strlen (httpviewer) + 10);
+			strcpy (tempbuf, httpviewer);
+			strcat (tempbuf, " ");
+			strcat (tempbuf, manuallinks[selected].name);
+			myendwin ();
+			system (tempbuf);
+			doupdate ();
+			xfree (tempbuf);
+		      }
+		    else if (!strncmp (manuallinks[selected].section, "FTPSECTION", 10))
+		      {
+			char *tempbuf = xmalloc (strlen (manuallinks[selected].name) + strlen (ftpviewer) + 10);
+			strcpy (tempbuf, httpviewer);
+			strcat (tempbuf, " ");
+			strcat (tempbuf, manuallinks[selected].name);
+			myendwin ();
+			system (tempbuf);
+			doupdate ();
+			xfree (tempbuf);
+		      }
+		    else if (!strncmp (manuallinks[selected].section, "MAILSECTION", 11))
+		      {
+			char *tempbuf = xmalloc (strlen (manuallinks[selected].name) + strlen (maileditor) + 10);
+			strcpy (tempbuf, maileditor);
+			strcat (tempbuf, " ");
+			strcat (tempbuf, manuallinks[selected].name);
+			myendwin ();
+			system ("clear");
+			system (tempbuf);
+			doupdate ();
+			xfree (tempbuf);
+		      }
+		    else
+		      {
+			return selected;
+		      }
+		  }
 	    }
 /*===========================================================================*/
 /*********************** end of keyboard handling ***************************/
@@ -753,9 +1122,9 @@ manualwork ()
 			{
 			  if (manuallinks[i].line == mouse.y + manualpos - 1)
 			    {
-			      if (manuallinks[i].row <= mouse.x - 1)
+			      if (manuallinks[i].col <= mouse.x - 1)
 				{
-				  if (manuallinks[i].row + strlen (manuallinks[i].name) >= mouse.x - 1)
+				  if (manuallinks[i].col + strlen (manuallinks[i].name) >= mouse.x - 1)
 				    {
 				      selected = i;
 				      done = 1;
@@ -769,9 +1138,9 @@ manualwork ()
 			  {
 			    if (manuallinks[i].line == mouse.y + manualpos - 1)
 			      {
-				if (manuallinks[i].row <= mouse.x - 1)
+				if (manuallinks[i].col <= mouse.x - 1)
 				  {
-				    if (manuallinks[i].row + strlen (manuallinks[i].name) >= mouse.x - 1)
+				    if (manuallinks[i].col + strlen (manuallinks[i].name) >= mouse.x - 1)
 				      {
 					selected = i;
 					done = 1;
@@ -794,9 +1163,9 @@ manualwork ()
 			{
 			  if (manuallinks[i].line == mouse.y + manualpos - 1)
 			    {
-			      if (manuallinks[i].row <= mouse.x - 1)
+			      if (manuallinks[i].col <= mouse.x - 1)
 				{
-				  if (manuallinks[i].row + strlen (manuallinks[i].name) >= mouse.x - 1)
+				  if (manuallinks[i].col + strlen (manuallinks[i].name) >= mouse.x - 1)
 				    {
 				      selected = i;
 				      done = 1;
@@ -810,9 +1179,9 @@ manualwork ()
 			  {
 			    if (manuallinks[i].line == mouse.y + manualpos - 1)
 			      {
-				if (manuallinks[i].row <= mouse.x - 1)
+				if (manuallinks[i].col <= mouse.x - 1)
 				  {
-				    if (manuallinks[i].row + strlen (manuallinks[i].name) >= mouse.x - 1)
+				    if (manuallinks[i].col + strlen (manuallinks[i].name) >= mouse.x - 1)
 				      {
 					selected = i;
 					done = 1;
@@ -833,15 +1202,23 @@ manualwork ()
 #endif
 /*****************************************************************************/
 	}
-      if (key == keys.quit_2)
-	break;
+      if ((key == keys.quit_2) || (key == keys.quit_1))
+	{
+	  if (!ConfirmQuit)
+	    break;
+	  else
+	    {
+	      if (yesno (_ ("Are you sure to quit?"), QuitConfirmDefault))
+		break;
+	    }
+	}
     }
   closeprogram ();
   return -1;
 }
 
 void
-rescan_selected ()
+rescan_selected ()		/* scan for some hyperlink, available on current screen */
 {
   int i;
   for (i = 0; i < ManualLinks; i++)
@@ -856,7 +1233,7 @@ rescan_selected ()
 }
 
 void
-showmanualscreen ()
+showmanualscreen ()		/* show the currently visible part of manpage */
 {
   int i;
 #ifdef getmaxyx
@@ -865,8 +1242,13 @@ showmanualscreen ()
   for (i = manualpos; (i < manualpos + (maxy - 2)) && (i < ManualLines); i++)
     /* print all visible text lines */
     {
+      int len = strlen (manual[i]);
+      if (len)
+	manual[i][len - 1] = ' ';
       mvaddstr_manual ((i - manualpos) + 1, 0, manual[i]);
-      clrtoeol ();
+      if (len)
+	manual[i][len - 1] = '\n';
+      myclrtoeol ();
     }
   clrtobot ();			/* and clear to bottom */
   add_highlights ();		/* add highlights */
@@ -874,7 +1256,7 @@ showmanualscreen ()
   mymvhline (0, 0, ' ', maxx);
   mymvhline (maxy - 1, 0, ' ', maxx);
   move (maxy - 1, 0);
-  if ((manualpos < ManualLines) && (ManualLines > maxy - 2))
+  if (((manualpos + maxy) < ManualLines) && (ManualLines > maxy - 2))
     printw (_ ("Viewing line %d/%d, %d%%"), (manualpos - 1 + maxy), ManualLines, ((manualpos - 1 + maxy) * 100) / ManualLines);
   else
     printw (_ ("Viewing line %d/%d, 100%%"), ManualLines, ManualLines);
@@ -927,7 +1309,7 @@ mvaddstr_manual (int y, int x, char *str)	/* print a manual line */
 	      addch (str[i] & A_CHARTEXT);
 	    }
 	}
-    label_skip_other:
+    label_skip_other:;
     }
 }
 void
@@ -953,7 +1335,34 @@ add_highlights ()		/* add hyperobject highlights */
 	      else
 		attrset (url);
 	    }
-	  mvaddstr (1 + manuallinks[i].line - manualpos, manuallinks[i].row, manuallinks[i].name);
+	  if (manuallinks[i].carry == 1)
+	    {
+	      int x, y, ltline = manuallinks[i].line - 1;
+	      char *tmpstr = strdup (manual[ltline]);
+	      int ltlinelen;
+	      char *newlinemark;
+	      strip_manual (tmpstr);
+	      ltlinelen = strlen (tmpstr);
+	      newlinemark = tmpstr + ltlinelen - 1;
+	      getyx (stdscr, y, x);
+	      if (y > 2)
+		{
+#define TestCh manual[ltline][ltlinelen]
+		  if (ltlinelen > 2)
+		    ltlinelen -= 3;
+		  while ((isalpha (TestCh)) || (TestCh == '.'))
+		    ltlinelen--;
+
+		  *newlinemark = 0;
+		  mvaddstr (manuallinks[i].line - manualpos + 1 - 1, ltlinelen,
+			    &tmpstr[ltlinelen]);
+		  *newlinemark = '\n';
+#undef TestCh
+		}
+	      xfree (tmpstr);
+	      move (y, x);
+	    }
+	  mvaddstr (1 + manuallinks[i].line - manualpos, manuallinks[i].col, manuallinks[i].name);
 	  attrset (normal);
 	}
     }
@@ -962,7 +1371,7 @@ void
 strip_manual (char *buf)	/* all variables passed here must have, say 10 bytes of overrun buffer */
 {
   int i, tmpcnt = 0;
-  for (i = 0; i < strlen (buf); i++)	/* in general, tmp buffer will hold a line stripped from highlight marks */
+  for (i = 0; buf[i] != 0; i++)	/* in general, tmp buffer will hold a line stripped from highlight marks */
     {
       if ((buf[i] == '_') && (buf[i + 1] == 8))		/* so we strip the line from "'_',0x8" -- bold marks */
 	{
@@ -981,7 +1390,9 @@ strip_manual (char *buf)	/* all variables passed here must have, say 10 bytes of
 }
 
 int
-is_in_manlinks (char *in, char *find)
+is_in_manlinks (char *in, char *find)	/* checks if a construction, which
+					 * looks like hyperlink, belongs
+					 * to the allowed manual sections.*/
 {
   char *copy, *token;
   const char delimiters[] = ":";
@@ -1006,3 +1417,18 @@ is_in_manlinks (char *in, char *find)
       return 0;
     }
 }
+
+void
+printmanual (char **Message, long Lines)
+{
+  FILE *prnFD;			/* printer fd */
+  int i;
+
+  prnFD = popen (printutility, "w");
+
+  for (i = 0; i < Lines; i++)	/* scan through all lines */
+    {
+      fprintf (prnFD, "\r%s", Message[i]);
+    }
+  pclose (prnFD);
+}
