@@ -1,3 +1,4 @@
+
 #include "common_includes.h"
 
 void
@@ -31,14 +32,14 @@ getnextnode (char *type, char *node)	/* read the `Next:' header entry */
   static char tmp[1024];
 #endif
   strcpy (tmp, type);
-  for (i = 0; i < strlen (tmp); i++)
+  for (i = 0; tmp[i] != 0; i++)
     {
       if (tmp[i] == ':')
 	{
 	  coloncount++;
 	  if (coloncount == 3)
 	    {
-	      for (j = i + 1; j < strlen (tmp); j++)
+	      for (j = i + 1; tmp[j] != 0; j++)
 		{
 		  if (tmp[j] == ',')
 		    {
@@ -68,14 +69,14 @@ getprevnode (char *type, char *node)	/* read the `Prev:' header entry */
   static char tmp[1024 + 1];
 #endif
   strcpy (tmp, type);
-  for (i = 0; i < strlen (tmp); i++)
+  for (i = 0; tmp[i] != 0; i++)
     {
       if (tmp[i] == ':')
 	{
 	  coloncount++;
 	  if (coloncount == 4)
 	    {
-	      for (j = i + 1; j < strlen (tmp); j++)
+	      for (j = i + 1; tmp[j] != 0; j++)
 		{
 		  if (tmp[j] == ',')
 		    {
@@ -105,14 +106,14 @@ getupnode (char *type, char *node)	/* read the `Up:' header entry */
   static char tmp[1024];
 #endif
   strcpy (tmp, type);
-  for (i = 0; i < strlen (tmp); i++)
+  for (i = 0; tmp[i] != 0; i++)
     {
       if (tmp[i] == ':')
 	{
 	  coloncount++;
 	  if (coloncount == 5)
 	    {
-	      for (j = i + 1; j < strlen (tmp); j++)
+	      for (j = i + 1; tmp[j] != 0; j++)
 		{
 		  if (tmp[j] == ',' || tmp[j] == '\n')
 		    {
