@@ -8,10 +8,12 @@ extern int CutManHeaders;
 extern int DontHandleWithoutTagTable;
 extern int use_manual;
 extern int use_raw_filename;
+extern int quote_ignored;
 extern char *httpviewer;
 extern char *ftpviewer;
 extern char *maileditor;
 extern char *manlinks;
+extern char *ignoredmacros;
 
 struct keybindings keys =
 {'s', 'S',			/* regexp search */
@@ -20,7 +22,7 @@ struct keybindings keys =
  'p', 'P',			/* previous node */
  'n', 'N',			/* next node */
  'u', 'U',			/* up node */
- KEY_UP, 'a',			/* up one line */
+ KEY_UP, 'k',			/* up one line */
 #ifdef HAVE_KEY_END
  KEY_END
 #else				/* HAVE_KEY_END */
@@ -28,11 +30,11 @@ struct keybindings keys =
 #endif				/* ~HAVE_KEY_END */
  ,'e',				/* end */
  KEY_NPAGE, ' ',		/* down one page */
- KEY_HOME, 'h',			/* home */
+ KEY_HOME, 'H',			/* home */
  KEY_PPAGE, '-',		/* up one page */
- KEY_DOWN, 'z',			/* down one line */
+ KEY_DOWN, 'j',			/* down one line */
  't', 'T',			/* top */
- KEY_LEFT, 'b',			/* back */
+ KEY_LEFT, 'h',			/* back */
  KEY_RIGHT, '\n',		/* follow link */
  'Q', 'q',			/* quit */
  12, '~',			/* refresh screen -- 12 is C-L */
@@ -40,29 +42,34 @@ struct keybindings keys =
  'D', 'd',			/* goto dir page */
  0, ' ',			/* pgdn_auto */
  0, '-',			/* pgup_auto */
- 'f',0				/* search again */
+ 'f', 0,			/* search again */
+ 'l', 0,			/* go to line */
+ KEY_IC, 0,			/* two lines up */
+ KEY_DC, 1,			/* two lines down */
+ ']', 0				/* print */
 };
 
 #ifndef NO_COLOR_CURSES
 struct colours cols =
-{COLOR_WHITE, COLOR_BLACK, NO_BOLD,	/* normal */
- COLOR_GREEN, COLOR_WHITE, BOLD,	/* selected menu */
- COLOR_GREEN, COLOR_BLACK, BOLD,	/* menu */
- COLOR_BLUE, COLOR_WHITE, BOLD,	/* selected note */
- COLOR_BLUE, COLOR_BLACK, BOLD,	/* note */
- COLOR_GREEN, COLOR_BLUE, BOLD,	/* top line */
- COLOR_GREEN, COLOR_BLUE, BOLD,	/* bottom line */
- COLOR_YELLOW, COLOR_BLACK, BOLD,	/* manual bold */
- COLOR_WHITE, COLOR_BLACK, BOLD,	/* manual italic */
- COLOR_MAGENTA, COLOR_BLACK, BOLD,	/* url */
- COLOR_MAGENTA, COLOR_GREEN, NO_BOLD,	/* url selected */
- COLOR_WHITE, COLOR_BLACK, BOLD	/* info highlight (quoted text) */
+{COLOR_WHITE, COLOR_BLACK, NO_BOLD, NO_BLINK,	/* normal */
+ COLOR_GREEN, COLOR_WHITE, BOLD, NO_BLINK,	/* selected menu */
+ COLOR_GREEN, COLOR_BLACK, BOLD, NO_BLINK,	/* menu */
+ COLOR_BLUE, COLOR_WHITE, BOLD, NO_BLINK,	/* selected note */
+ COLOR_BLUE, COLOR_BLACK, BOLD, NO_BLINK,	/* note */
+ COLOR_GREEN, COLOR_BLUE, BOLD, NO_BLINK,	/* top line */
+ COLOR_GREEN, COLOR_BLUE, BOLD, NO_BLINK,	/* bottom line */
+ COLOR_YELLOW, COLOR_BLACK, BOLD, NO_BLINK,	/* manual bold */
+ COLOR_WHITE, COLOR_BLACK, BOLD, NO_BLINK,	/* manual italic */
+ COLOR_MAGENTA, COLOR_BLACK, BOLD, NO_BLINK,	/* url */
+ COLOR_MAGENTA, COLOR_GREEN, NO_BOLD, NO_BLINK,		/* url selected */
+ COLOR_WHITE, COLOR_BLACK, BOLD, NO_BLINK,	/* info highlight (quoted text) */
+ COLOR_YELLOW, COLOR_BLACK, BOLD, NO_BLINK	/* search highlight */
 };
 #endif /* NO_COLOR_CURSES */
 int
 parse_config (void)
 {
-  char config_file_name[256], *home;
+  char config_file_name[256], *home = 0;
   char line[256];
   FILE *f;
   int line_number = 0;
@@ -76,14 +83,22 @@ parse_config (void)
 	{
 	  strcpy (config_file_name, CONFIGDIR);
 	  if (!(f = fopen (config_file_name, "r")))
-	    return 0;		/* no config file available */
+	    {
+	      if (home)
+		free (home);
+	      return 0;		/* no config file available */
+	    }
 	}
     }
   else
     {
       strcpy (config_file_name, CONFIGDIR);
       if (!(f = fopen (config_file_name, "r")))
-	return 0;
+	{
+	  if (home)
+	    free (home);
+	  return 0;
+	}
     }
 
   while (!feof (f))
@@ -91,6 +106,8 @@ parse_config (void)
       if (!(fgets (line, 255, f)))
 	{
 	  fclose (f);
+	  if (home)
+	    free (home);
 	  return 0;
 	}
       if (parse_line (line))
@@ -105,6 +122,8 @@ parse_config (void)
     }
 
   fclose (f);
+  if (home)
+    free (home);
   return 0;
 }
 
@@ -114,16 +133,18 @@ parse_line (char *line)
   char *temp;
   int *fore = NULL;
   int *key = NULL;
-  char *tmp = line;
 #ifdef HAS_CURSES
-  int *back = NULL, *bold = NULL, *p = NULL;
+  int *back = NULL, *bold = NULL, *blink = NULL, *p = NULL;
   int i;
 #endif /* HAS_CURSES */
 
   if (line[0] == '#')
     return 0;
 
-  temp = str_toupper (skip_whitespace (strtok (line, "=")));
+  if (!(temp = skip_whitespace (strtok (line, "="))))
+    return 1;
+
+  temp = str_toupper (temp);
 
   if (strlen (temp) < 1)
     return 0;
@@ -171,6 +192,10 @@ parse_line (char *line)
 	key = &keys.up_1;
       else if (!strncmp (temp + 4, "UP_2", 4))
 	key = &keys.up_2;
+      else if (!strncmp (temp + 4, "TWOUP_1", 7))
+	key = &keys.twoup_1;
+      else if (!strncmp (temp + 4, "TWOUP_2", 7))
+	key = &keys.twoup_2;
       else if (!strncmp (temp + 4, "END_1", 5))
 	key = &keys.end_1;
       else if (!strncmp (temp + 4, "END_2", 5))
@@ -191,6 +216,10 @@ parse_line (char *line)
 	key = &keys.down_1;
       else if (!strncmp (temp + 4, "DOWN_2", 6))
 	key = &keys.down_2;
+      else if (!strncmp (temp + 4, "TWODOWN_1", 9))
+	key = &keys.twodown_1;
+      else if (!strncmp (temp + 4, "TWODOWN_2", 9))
+	key = &keys.twodown_2;
       else if (!strncmp (temp + 4, "TOP_1", 5))
 	key = &keys.top_1;
       else if (!strncmp (temp + 4, "TOP_2", 5))
@@ -219,6 +248,14 @@ parse_line (char *line)
 	key = &keys.dirpage_1;
       else if (!strncmp (temp + 4, "DIRPAGE_2", 9))
 	key = &keys.dirpage_2;
+      else if (!strncmp (temp + 4, "GOLINE_1", 8))
+	key = &keys.goline_1;
+      else if (!strncmp (temp + 4, "GOLINE_2", 8))
+	key = &keys.goline_2;
+      else if (!strncmp (temp + 4, "PRINT_1", 7))
+	key = &keys.print_1;
+      else if (!strncmp (temp + 4, "PRINT_2", 7))
+	key = &keys.print_2;
       else
 	return 1;
     }
@@ -231,72 +268,91 @@ parse_line (char *line)
 	  fore = &cols.normal_fore;
 	  back = &cols.normal_back;
 	  bold = &cols.normal_bold;
+	  blink = &cols.normal_blink;
 	}
       else if (!strncmp (temp + 4, "MENUSELECTED", 12))
 	{
 	  fore = &cols.menuselected_fore;
 	  back = &cols.menuselected_back;
 	  bold = &cols.menuselected_bold;
+	  blink = &cols.menuselected_blink;
 	}
       else if (!strncmp (temp + 4, "MENU", 4))
 	{
 	  fore = &cols.menu_fore;
 	  back = &cols.menu_back;
 	  bold = &cols.menu_bold;
+	  blink = &cols.menu_blink;
 	}
       else if (!strncmp (temp + 4, "NOTESELECTED", 12))
 	{
 	  fore = &cols.noteselected_fore;
 	  back = &cols.noteselected_back;
 	  bold = &cols.noteselected_bold;
+	  blink = &cols.noteselected_blink;
 	}
       else if (!strncmp (temp + 4, "NOTE", 4))
 	{
 	  fore = &cols.note_fore;
 	  back = &cols.note_back;
 	  bold = &cols.note_bold;
+	  blink = &cols.note_blink;
 	}
       else if (!strncmp (temp + 4, "TOPLINE", 7))
 	{
 	  fore = &cols.topline_fore;
 	  back = &cols.topline_back;
 	  bold = &cols.topline_bold;
+	  blink = &cols.topline_blink;
 	}
       else if (!strncmp (temp + 4, "BOTTOMLINE", 10))
 	{
 	  fore = &cols.bottomline_fore;
 	  back = &cols.bottomline_back;
 	  bold = &cols.bottomline_bold;
+	  blink = &cols.bottomline_blink;
 	}
       else if (!strncmp (temp + 4, "MANUALBOLD", 10))
 	{
 	  fore = &cols.manualbold_fore;
 	  back = &cols.manualbold_back;
 	  bold = &cols.manualbold_bold;
+	  blink = &cols.manualbold_blink;
 	}
       else if (!strncmp (temp + 4, "MANUALITALIC", 12))
 	{
 	  fore = &cols.manualitalic_fore;
 	  back = &cols.manualitalic_back;
 	  bold = &cols.manualitalic_bold;
+	  blink = &cols.manualitalic_blink;
 	}
       else if (!strncmp (temp + 4, "URLSELECTED", 11))
 	{
 	  fore = &cols.urlselected_fore;
 	  back = &cols.urlselected_back;
 	  bold = &cols.urlselected_bold;
+	  blink = &cols.urlselected_blink;
 	}
       else if (!strncmp (temp + 4, "URL", 3))
 	{
 	  fore = &cols.url_fore;
 	  back = &cols.url_back;
 	  bold = &cols.url_bold;
+	  blink = &cols.url_blink;
 	}
       else if (!strncmp (temp + 4, "INFOHIGHLIGHT", 13))
 	{
 	  fore = &cols.infohighlight_fore;
 	  back = &cols.infohighlight_back;
 	  bold = &cols.infohighlight_bold;
+	  blink = &cols.infohighlight_blink;
+	}
+      else if (!strncmp (temp + 4, "SEARCHHIGHLIGHT", 13))
+	{
+	  fore = &cols.searchhighlight_fore;
+	  back = &cols.searchhighlight_back;
+	  bold = &cols.searchhighlight_bold;
+	  blink = &cols.searchhighlight_blink;
 	}
       else
 	return 1;
@@ -307,7 +363,8 @@ parse_line (char *line)
       temp = strtok (NULL, "=");
       if (temp)
 	{
-	  temp = str_toupper (skip_whitespace (temp));
+	  if (!(temp = str_toupper (skip_whitespace (temp))))
+	    return 1;
 	  if (!strncmp (temp, "TRUE", 4))
 	    use_manual = 1;
 	  else if (!strncmp (temp, "FALSE", 5))
@@ -321,7 +378,8 @@ parse_line (char *line)
       temp = strtok (NULL, "=");
       if (temp)
 	{
-	  temp = str_toupper (skip_whitespace (temp));
+	  if (!(temp = str_toupper (skip_whitespace (temp))))
+	    return 1;
 	  if (!strncmp (temp, "TRUE", 4))
 	    use_raw_filename = 1;
 	  else if (!strncmp (temp, "FALSE", 5))
@@ -335,7 +393,8 @@ parse_line (char *line)
       temp = strtok (NULL, "=");
       if (temp)
 	{
-	  temp = str_toupper (skip_whitespace (temp));
+	  if (!(temp = str_toupper (skip_whitespace (temp))))
+	    return 1;
 	  if (!strncmp (temp, "TRUE", 4))
 	    use_apropos = 1;
 	  else if (!strncmp (temp, "FALSE", 5))
@@ -344,12 +403,43 @@ parse_line (char *line)
 	    return 1;
 	}
     }
+  else if (!strncmp (temp, "QUIT-CONFIRMATION", 17))
+    {
+      temp = strtok (NULL, "=");
+      if (temp)
+	{
+	  if (!(temp = str_toupper (skip_whitespace (temp))))
+	    return 1;
+	  if (!strncmp (temp, "TRUE", 4))
+	    ConfirmQuit = 1;
+	  else if (!strncmp (temp, "FALSE", 5))
+	    ConfirmQuit = 0;
+	  else
+	    return 1;
+	}
+    }
+  else if (!strncmp (temp, "QUIT-CONFIRM-DEFAULT", 20))
+    {
+      temp = strtok (NULL, "=");
+      if (temp)
+	{
+	  if (!(temp = str_toupper (skip_whitespace (temp))))
+	    return 1;
+	  if (!strncmp (temp, "YES", 3))
+	    QuitConfirmDefault = 1;
+	  else if (!strncmp (temp, "NO", 2))
+	    QuitConfirmDefault = 0;
+	  else
+	    return 1;
+	}
+    }
   else if (!strncmp (temp, "CUT-MAN-HEADERS", 15))
     {
       temp = strtok (NULL, "=");
       if (temp)
 	{
-	  temp = str_toupper (skip_whitespace (temp));
+	  if (!(temp = str_toupper (skip_whitespace (temp))))
+	    return 1;
 	  if (!strncmp (temp, "TRUE", 4))
 	    CutManHeaders = 1;
 	  else if (!strncmp (temp, "FALSE", 5))
@@ -358,12 +448,28 @@ parse_line (char *line)
 	    return 1;
 	}
     }
+  else if (!strncmp (temp, "CLEAR-SCREEN-AT-EXIT", 20))
+    {
+      temp = strtok (NULL, "=");
+      if (temp)
+	{
+	  if (!(temp = str_toupper (skip_whitespace (temp))))
+	    return 1;
+	  if (!strncmp (temp, "TRUE", 4))
+	    ClearScreenAtExit = 1;
+	  else if (!strncmp (temp, "FALSE", 5))
+	    ClearScreenAtExit = 0;
+	  else
+	    return 1;
+	}
+    }
   else if (!strncmp (temp, "CUT-EMPTY-MAN-LINES", 19))
     {
       temp = strtok (NULL, "=");
       if (temp)
 	{
-	  temp = str_toupper (skip_whitespace (temp));
+	  if (!(temp = str_toupper (skip_whitespace (temp))))
+	    return 1;
 	  if (!strncmp (temp, "TRUE", 4))
 	    CutEmptyManLines = 1;
 	  else if (!strncmp (temp, "FALSE", 5))
@@ -377,7 +483,8 @@ parse_line (char *line)
       temp = strtok (NULL, "=");
       if (temp)
 	{
-	  temp = str_toupper (skip_whitespace (temp));
+	  if (!(temp = str_toupper (skip_whitespace (temp))))
+	    return 1;
 	  if (!strncmp (temp, "TRUE", 4))
 	    DontHandleWithoutTagTable = 1;
 	  else if (!strncmp (temp, "FALSE", 5))
@@ -391,7 +498,8 @@ parse_line (char *line)
       temp = strtok (NULL, "=");
       if (temp)
 	{
-	  temp = str_toupper (skip_whitespace (temp));
+	  if (!(temp = str_toupper (skip_whitespace (temp))))
+	    return 1;
 	  if (!strncmp (temp, "TRUE", 4))
 	    LongManualLinks = 1;
 	  else if (!strncmp (temp, "FALSE", 5))
@@ -433,6 +541,17 @@ parse_line (char *line)
       else
 	return 1;
     }
+  else if (!strncmp (temp, "PRINTUTILITY", 12))
+    {
+      temp = strtok (NULL, "\n");
+      if (temp)
+	{
+	  printutility = strdup (temp);
+	  remove_quotes (printutility);
+	}
+      else
+	return 1;
+    }
   else if (!strncmp (temp, "MAN-OPTIONS", 11))
     {
       temp = strtok (NULL, "\n");
@@ -455,6 +574,21 @@ parse_line (char *line)
       else
 	return 1;
     }
+  else if (!strncmp (temp, "FILTER-0XB7", 11))
+    {
+      temp = strtok (NULL, "=");
+      if (temp)
+	{
+	  if (!(temp = str_toupper (skip_whitespace (temp))))
+	    return 1;
+	  if (!strncmp (temp, "TRUE", 4))
+	    FilterB7 = 1;
+	  else if (!strncmp (temp, "FALSE", 5))
+	    FilterB7 = 0;
+	  else
+	    return 1;
+	}
+    }
   else if (!strncmp (temp, "MANLINKS", 8))
     {
       temp = strtok (NULL, "\n");
@@ -466,21 +600,57 @@ parse_line (char *line)
       else
 	return 1;
     }
+  else if (!strncmp (temp, "QUOTE-IGNORED-MACROS", 20))
+    {
+      temp = strtok (NULL, "=");
+      if (temp)
+	{
+	  if (!(temp = str_toupper (skip_whitespace (temp))))
+	    return 1;
+	  if (!strncmp (temp, "TRUE", 4))
+	    quote_ignored = 1;
+	  else if (!strncmp (temp, "FALSE", 5))
+	    quote_ignored = 0;
+	  else
+	    return 1;
+	}
+    }
+
+  else if (!strncmp (temp, "IGNORE-MACROS", 8))
+    {
+      temp = strtok (NULL, "\n");
+      if (temp)
+	{
+	  ignoredmacros = strdup (temp);
+	  remove_quotes (ignoredmacros);
+	  if (ignoredmacros[0] == '\t' || ignoredmacros[0] == ' '
+	      || !strncasecmp (ignoredmacros, "FALSE", 5))
+	    ignoredmacros[0] = '\0';
+	}
+      else
+	return 1;
+    }
   else
     return 1;
 #ifndef NO_COLOR_CURSES
   if (fore)
     {
-      for (i = 0; i < 3; i++)
+      for (i = 0; i < 4; i++)
 	{
 	  if (i == 0)
 	    p = fore;
 	  else if (i == 1)
 	    p = back;
-	  else
+	  else if (i == 2)
 	    p = bold;
+	  else
+	    p = blink;
+
+	  if (!(temp = skip_whitespace (strtok (NULL, ","))))
+	    return 1;
+
+	  temp = str_toupper (temp);
 
-	  temp = str_toupper (skip_whitespace (strtok (NULL, ",")));
 	  if (!(strncmp (temp, "COLOR_BLACK", 11)))
 	    *p = COLOR_BLACK;
 	  else if (!(strncmp (temp, "COLOR_RED", 9)))
@@ -503,6 +673,10 @@ parse_line (char *line)
 	    *p = BOLD;
 	  else if (!(strncmp (temp, "NO_BOLD", 7)))
 	    *p = 0;
+	  else if (!(strncmp (temp, "BLINK", 4)))
+	    *p = BOLD;
+	  else if (!(strncmp (temp, "NO_BLINK", 7)))
+	    *p = 0;
 	  else
 	    return 1;
 	}
@@ -511,7 +685,8 @@ parse_line (char *line)
 #endif /* NO_COLOR_CURSES */
   if (key)
     {
-      temp = skip_whitespace (strtok (NULL, "="));
+      if (!(temp = skip_whitespace (strtok (NULL, "="))))
+	return 0;
       if (!(strncmp (temp, "KEY_", 4)) ||
 	  !(strncmp (temp, "key_", 4)))
 	{
@@ -527,6 +702,10 @@ parse_line (char *line)
 	    *key = KEY_LEFT;
 	  else if (!(strncmp (temp + 4, "RIGHT", 5)))
 	    *key = KEY_RIGHT;
+	  else if (!(strncmp (temp + 4, "IC", 2)))
+	    *key = KEY_IC;
+	  else if (!(strncmp (temp + 4, "DC", 2)))
+	    *key = KEY_DC;
 	  else if (!(strncmp (temp + 4, "HOME", 4)))
 	    *key = KEY_HOME;
 	  else if (!(strncmp (temp + 4, "BACKSPACE", 9)))
@@ -565,15 +744,16 @@ parse_line (char *line)
 #endif
 	  else if (!(strncmp (temp + 4, "CTRL", 4)))
 	    {
-	      temp = skip_whitespace (temp + 8);
+	      if (!(temp = skip_whitespace (temp + 8)))
+		return 1;
 	      if (temp[0] == '(')
 		{
 		  if (temp[1] == '\'')
-		    *key = KEY_CTRL(temp[2]);
+		    *key = KEY_CTRL (temp[2]);
 		  else if (isdigit (temp[1]))
 		    {
 		      char *tail = temp + (strlen (temp));
-		      *key = KEY_CTRL((int) strtol (temp + 1, &tail, 10));
+		      *key = KEY_CTRL ((int) strtol (temp + 1, &tail, 10));
 		    }
 		  else
 		    return 1;
@@ -583,15 +763,16 @@ parse_line (char *line)
 	    }
 	  else if (!(strncmp (temp + 4, "ALT", 3)))
 	    {
-	      temp = skip_whitespace (temp + 7);
+	      if (!(temp = skip_whitespace (temp + 7)))
+		return 1;
 	      if (temp[0] == '(')
 		{
 		  if (temp[1] == '\'')
-		    *key = KEY_ALT(tolower(temp[2]));
+		    *key = KEY_ALT (tolower (temp[2]));
 		  else if (isdigit (temp[1]))
 		    {
 		      char *tail = temp + (strlen (temp));
-		      *key = KEY_ALT((int) strtol (temp + 1, &tail, 10));
+		      *key = KEY_ALT ((int) strtol (temp + 1, &tail, 10));
 		    }
 		  else
 		    return 1;
@@ -599,8 +780,6 @@ parse_line (char *line)
 	      else
 		return 1;
 	    }
-	  
-	  
 	  else
 	    return 1;
 	}
@@ -651,6 +830,9 @@ skip_whitespace (char *str)
 {
   int i = 0;
 
+  if (!str)
+    return NULL;
+
   while (str[i] == ' ' || str[i] == '\t' || str[i] == '\n')
     i++;
 
