@@ -9,6 +9,9 @@
 #include <getopt.h>
 #endif
 
+#define	SAFE_USER	"nobody"
+#define	SAFE_GROUP	SAFE_USER
+
 char *version = VERSION;
 int DontHandleWithoutTagTable = 0;
 
@@ -22,6 +25,8 @@ void strip_file_from_info_suffix (char *file);	/*
 						 */
 char *addinfosuffix (char *file);	/* add `.info' suffix to "file" */
 
+void checksu ();		/* protect against bad, bad macros */
+
 int
 main (int argc, char *argv[])
 {
@@ -35,7 +40,6 @@ main (int argc, char *argv[])
   long lines = 0;		/* line count in message */
   char **message = 0;		/* this will hold node's text */
   char *type = 0;		/* this will hold the node's header */
-  char *tmp;
   int tag_table_pos = 1;
 #ifdef HAVE_GETOPT_LONG
   static struct option long_options[] =
@@ -46,16 +50,18 @@ main (int argc, char *argv[])
     {"raw-filename", 0, 0, 'r'},
     {"apropos", 0, 0, 'a'},
     {"cut-man-headers", 0, 0, 'c'},
-    {"empty-lines", 0, 0, 'e'},
+    {"squeeze-manlines", 0, 0, 's'},
     {"dont-handle-without-tag-table", 0, 0, 'd'},
     {"force-manual-tag-table", 0, 0, 't'},
     {"node", 1, 0, 'n'},
     {"long-manual-links", 0, 0, 'l'},
+    {"clear-at-exit", 0, 0, 'x'},
     {0, 0, 0, 0}};
 #endif
   signal_handler ();		/* take care of SIGSEGV, SIGTERM, SIGINT */
-  searchagain.type=0;
-  searchagain.search=0;
+  searchagain.type = 0;
+  searchagain.search = 0;
+  checksu ();
   initlocale ();
   inithistory ();
   printf ("Przemek's Info Viewer v%s\n", version);
@@ -66,6 +72,7 @@ main (int argc, char *argv[])
       id = openinfo ("dir", 0);
       curfile = xmalloc (150);
       strcpy (curfile, "dir");
+      strcpy (filename, "dir");
     }
   if (strlen (argv[0]) >= 3)
     if (strstr (argv[0], "man") != NULL)	/* handle any 'man' alias to 'pinfo' */
@@ -88,20 +95,23 @@ main (int argc, char *argv[])
 
   do
     {
-      command_line_option = getopt_long (argc, argv, "hvmracdetnl", long_options, NULL);
+      command_line_option = getopt_long (argc, argv, "hvmracsdtnlx", long_options, NULL);
       switch (command_line_option)
 	{
+	case 'x':
+	  ClearScreenAtExit = 1;
+	  break;
 	case 'l':
-	  LongManualLinks=1;
+	  LongManualLinks = 1;
 	  break;
 	case 'n':
-	  if(!optarg)
+	  if (!optarg)
 	    {
-	      printf(_("--node option used without argument\n"));
-	      exit(1);
+	      printf (_ ("--node option used without argument\n"));
+	      exit (1);
 	    }
-	  pinfo_start_node=malloc(strlen(optarg)+1);
-	  strcpy(pinfo_start_node,optarg);
+	  pinfo_start_node = malloc (strlen (optarg) + 1);
+	  strcpy (pinfo_start_node, optarg);
 	  break;
 	case 't':
 	  ForceManualTagTable = 1;
@@ -110,11 +120,19 @@ main (int argc, char *argv[])
 	  printf (_ ("Usage:\n" \
 		     "%s [options] [info|manual]\n" \
 		     "Options:\n" \
-		     "--help, -h: help\n" \
-		     "--version, -v: version\n" \
-		     "--manual, -m: use man page\n" \
-		     "--raw-filename, -r: use raw filename\n" \
-		     "--apropos, -a: call apropos if nothing found\n"),
+		     "-h, --help                            help\n" \
+		     "-v, --version                         version\n" \
+		     "-m, --manual                          use man page\n" \
+		"-r, --raw-filename                    use raw filename\n" \
+		     "-a, --apropos                         call apropos if nothing found\n" \
+		     "-c, --cut-man-headers                 cut out repeated man headers\n" \
+		     "-l, --long-manual-links               use long link names in manuals\n" \
+		     "-s, --squeeze-manlines                cut empty lines from manual pages\n" \
+		     "-d, --dont-handle-without-tag-table   don't display texinfo pages without tag\n" \
+		     "                                      tables\n" \
+		     "-t, --force-manual-tag-table          force manual detection of tag table\n" \
+	    "-x, --clear-at-exit                   clear screen at exit\n" \
+		     "    --node=nodename, --node nodename  jump directly to the node nodename\n"),
 		  argv[0]);
 	  exit (0);
 	case 'v':
@@ -131,6 +149,7 @@ main (int argc, char *argv[])
 	case 'r':
 	  strncpy (filename, argv[argc - 1], 200);
 	  checkfilename (filename);	/* security check */
+	  strip_compression_suffix (filename);	/* later, openinfo automaticaly adds them */
 	  id = openinfo (filename, 0);	/* try without '.info' suffix */
 	  break;
 	case 'a':
@@ -142,7 +161,7 @@ main (int argc, char *argv[])
 	case 'd':
 	  DontHandleWithoutTagTable = 1;
 	  break;
-	case 'e':
+	case 's':
 	  CutEmptyManLines = 1;
 	  break;
 	}
@@ -153,8 +172,21 @@ main (int argc, char *argv[])
 
   if (argc > 1)
     {
+#ifdef HAVE_GETOPT_LONG
+      if (optind < argc)
+	{
+	  strncpy (filename, argv[optind], 200);	/* the paths will be searched
+							   by openinfo() */
+	}
+      else
+	{
+	  strcpy (filename, "dir");
+	}
+
+#else
       strncpy (filename, argv[argc - 1], 200);	/* the paths will be searched
 						   by openinfo() */
+#endif
       checkfilename (filename);	/* security check */
 
       curfile = xmalloc (strlen (filename) + 100);	/* leave some space for
@@ -167,10 +199,25 @@ main (int argc, char *argv[])
       strcat (filename, ".info");
       id = openinfo (filename, 0);
     }
+  if (id == NULL)		/* try to lookup the name in dir file */
+    {
+      id = dirpage_lookup (&type, &message, &lines, filename, &pinfo_start_node);
+    }
   if (id == NULL)		/* if still nothing, try to use man page instead */
     {
       printf (_ ("Error: could not open info file, trying manual\n"));
+#ifdef HAVE_GETOPT_LONG
+      if (optind < argc)
+	{
+	  strncpy (filename, argv[optind], 200);	/* security check */
+	}
+      else
+	{
+	  strcpy (filename, "dir");
+	}
+#else
       strncpy (filename, argv[argc - 1], 200);	/* security check */
+#endif
       exit (handlemanual (filename));
     }
   if (seek_indirect (id))	/* search for indirect entries, if any */
@@ -216,17 +263,20 @@ main (int argc, char *argv[])
       else
 	return 1;
     }
-    
-  if(pinfo_start_node)
+
+  if (pinfo_start_node)
     {
-      tag_table_pos=gettagtablepos(pinfo_start_node);
-      if(tag_table_pos==-1)
-        {
-          printf(_("Specified node does not exist...\n"));
-          return 1;
-        }
+      tag_table_pos = gettagtablepos (pinfo_start_node);
+      if (tag_table_pos == -1)
+	{
+	  printf (_ ("Specified node does not exist...\n"));
+	  return 1;
+	}
+    }
+  else
+    {
+      tag_table_pos = gettagtablepos (FirstNodeName);
     }
-    
   init_curses ();		/* initialize curses screen interface */
 
   do
@@ -353,10 +403,10 @@ main (int argc, char *argv[])
 			  if (tmptagtablepos != -1)
 			    tag_table_pos = tmptagtablepos;
 			  else
-			    tag_table_pos = 1;
+			    tag_table_pos = gettagtablepos (FirstNodeName);
 			}
 		      else
-			tag_table_pos = 1;
+			tag_table_pos = gettagtablepos (FirstNodeName);
 
 		    }		/* end: open new info file -- file exists */
 		}		/* end: open new info file */
@@ -399,3 +449,49 @@ addinfosuffix (char *info)
 
   return withsuffix;
 }
+
+/* If pinfo was called by root then it should work as nobody.   */
+/* This protect us against .pso and .open macros which could    */
+/* be used for breaking the system's security.                  */
+
+void
+checksu ()
+{
+  struct passwd *pswd;
+  struct group *grwd;
+
+  if (!getegid () || !getgid ())
+    {
+      grwd = getgrnam (SAFE_GROUP);
+      if (!grwd)
+	{
+	  printf (_ ("Security warning: Unable to get GID of group called: %s\n"), SAFE_GROUP);
+	  sleep (1);
+	}
+      else
+	{
+	  if (!getgid () && !getuid ())
+	    setgid (grwd->gr_gid);
+	  else
+	    setegid (grwd->gr_gid);
+	}
+    }
+
+  if (!geteuid () || !getuid ())
+    {
+      pswd = getpwnam (SAFE_USER);
+      if (!pswd)
+	{
+	  printf (_ ("Security warning: Unable to get UID of user called: %s\n"), SAFE_USER);
+	  sleep (1);
+	}
+      else
+	{
+	  if (!getuid ())
+	    setuid (pswd->pw_uid);
+	  else
+	    seteuid (pswd->pw_uid);
+	}
+    }
+
+}
