@@ -47,7 +47,7 @@ readlinewrapper (char *prompt)
 	case KEY_BACKSPACE:	/* chars starting from curpos */
 	  if (cursor > 0)	/* - 1 from buf[n+1] to buf   */
 	    {
-	      for (i = cursor - 1; i < strlen (buf); i++)
+	      for (i = cursor - 1; buf[i] != 0; i++)
 		buf[i] = buf[i + 1];
 	      cursor--;
 	    }
@@ -55,7 +55,7 @@ readlinewrapper (char *prompt)
 	case KEY_DC:		/* handle delete key. As above */
 	  if (cursor < strlen (buf) - 1)
 	    {
-	      for (i = cursor; i < strlen (buf); i++)
+	      for (i = cursor; buf[i] != 0; i++)
 		buf[i] = buf[i + 1];
 	      cursor--;
 	    }
@@ -141,6 +141,7 @@ readlinewrapper (char *prompt)
     }
   else
     {
+      xfree (rlhistory[rlhistorylen - 1]);
       rlhistorylen--;
       rlhistorypos = rlhistorylen;
     }
