@@ -6,8 +6,9 @@ void
 handle_crash (int signum)
 {
   closeprogram ();
-  fprintf (stderr, "catched signal %d\n", signum);
-  perror ("pinfo: crash with");
+  fprintf (stderr, "Catched signal %d, bye!\n", signum);
+  if (signum == SIGSEGV)
+    perror ("pinfo: crash with");
   exit (1);
 }
 
@@ -16,7 +17,7 @@ handle_window_resize (int signum)
 {
 #ifdef	USE_NCURSES
   struct winsize winsz;
-  endwin ();
+  myendwin ();
   ioctl (0, TIOCGWINSZ, &winsz);
   COLS = winsz.ws_col;
   LINES = winsz.ws_row;
