@@ -52,17 +52,20 @@ void
 showscreen (char **message, char *type, long lines, long pos, long cursor)
 {
   long i;
-  for (i = pos; (i < lines - 1) && (i < pos + maxy - 1); i++)
+  for (i = pos; (i < lines) && (i < pos + maxy - 2); i++)
     {
+      int tmp = strlen (message[i]) - 1;
+      message[i][tmp] = 0;
       mvaddstr (i + 1 - pos, 0, message[i]);
-      clrtoeol ();
+      myclrtoeol ();
+      message[i][tmp] = '\n';
     }
   clrtobot ();
   attrset (bottomline);
   mymvhline (maxy - 1, 0, ' ', maxx);
   move (maxy - 1, 0);
-  if ((pos < lines - 1) && (lines > maxy - 2))
-    printw (_ ("Viewing line %d/%d, %d%%"), pos + maxy - 1, lines, ((pos + maxy - 1) * 100) / lines);
+  if ((pos < lines - 1) && (lines > pos + maxy - 2))
+    printw (_ ("Viewing line %d/%d, %d%%"), pos + maxy - 2, lines, ((pos + maxy - 2) * 100) / lines);
   else
     printw (_ ("Viewing line %d/%d, 100%%"), lines, lines);
   info_add_highlights (pos, cursor);
@@ -110,7 +113,7 @@ info_add_highlights (int pos, int cursor)
 	      if (hyperobjects[i].breakpos == -1)
 		{
 		  mvaddstr (1 + hyperobjects[i].line - pos,
-			    hyperobjects[i].row,
+			    hyperobjects[i].col,
 			    hyperobjects[i].node);
 		}
 	      else
@@ -118,7 +121,7 @@ info_add_highlights (int pos, int cursor)
 		  char tmp = hyperobjects[i].node[hyperobjects[i].breakpos];
 		  hyperobjects[i].node[hyperobjects[i].breakpos] = 0;
 		  mvaddstr (1 + hyperobjects[i].line - pos,
-			    hyperobjects[i].row,
+			    hyperobjects[i].col,
 			    hyperobjects[i].node);
 		  hyperobjects[i].node[hyperobjects[i].breakpos] = tmp;
 		  if (hyperobjects[i].line - pos + 3 < maxy)
@@ -132,7 +135,7 @@ info_add_highlights (int pos, int cursor)
 	      if (hyperobjects[i].breakpos == -1)
 		{
 		  move (1 + hyperobjects[i].line - pos,
-			hyperobjects[i].row);
+			hyperobjects[i].col);
 		  addstr ("(");
 		  addstr (hyperobjects[i].file);
 		  addstr (")");
@@ -149,7 +152,7 @@ info_add_highlights (int pos, int cursor)
 		  tmp = buf[hyperobjects[i].breakpos];
 		  buf[hyperobjects[i].breakpos] = 0;
 		  mvaddstr (1 + hyperobjects[i].line - pos,
-			    hyperobjects[i].row,
+			    hyperobjects[i].col,
 			    buf);
 		  buf[hyperobjects[i].breakpos] = tmp;
 		  if (hyperobjects[i].line - pos + 3 < maxy)
